# 2020-04-17

# Kiyoto Aramis Tanemura

# Code for Louvain algorithm was getting lengthy. I dedicate its own module

import numpy as np

from .centroid import centroid_autograph

def getModularity(affinityMatrix, communityAssignments, resolution = 1.0):
    # Provided an affinity matrix (np.array) and nodes specifying their assigned communities (list),
    # return the modularity of the whole graph
    arrayDim = affinityMatrix.shape[0]
    communities = list(set(communityAssignments))
    Q = 0
    affinityMatrix[range(arrayDim), range(arrayDim)] += affinityMatrix[range(arrayDim), range(arrayDim)]
    two_m = np.sum(affinityMatrix)
    for C in communities:
        communityIndices = [i for i in range(len(communityAssignments)) if communityAssignments[i] == C]
        sigma_in = np.sum(affinityMatrix[communityIndices, :][:, communityIndices])
        sigma_tot = np.sum(affinityMatrix[communityIndices, :])
        Q += sigma_in / two_m - resolution * (sigma_tot / two_m) ** 2
    return Q

def LouvainPhase1(affinityMatrix, communityAssignment, Q_threshold, max_iter, resolution):
    # inputs: affinityMatrix as numpy array. Clear diagonals prior to entering and filter edges with weights below threshold
    # communityAssignments: list of length of nodes. Values are community IDs
    # Q_threshold: minimum global modularity to terminate phase 1
    # max_iter: number of iteration to force termination of phase 1
    # output: updated community Assignment list

    num_nodes = affinityMatrix.shape[0]

    affinityMatrix[range(num_nodes), range(num_nodes)] += affinityMatrix[range(num_nodes), range(num_nodes)]

    two_m = np.sum(affinityMatrix)

    modularity_current = getModularity(affinityMatrix, communityAssignment, resolution)
    changeModularity = 1

    adjacencyMatrix = affinityMatrix > 0
    iterations = 0

    while changeModularity > Q_threshold and iterations < max_iter:
        modularity_prev = modularity_current
        for i in range(num_nodes): # for each node
            neighborIndices = [j for j in range(num_nodes) if adjacencyMatrix[i,j] == 1]
            communities = set([communityAssignment[k] for k in neighborIndices])
            communities.discard(communityAssignment[i])
            communities = list(communities)
            modularityList = []
            # ki: sum of edges incident to node i
            ki = np.sum(affinityMatrix[i, :])
            for C in communities:
                C_members = [x for x in range(num_nodes) if communityAssignment[x] == C]
                # Sum weights of all edges incident to nodes in C
                sigma_tot = np.sum(affinityMatrix[C_members, :])
                # ki_in: sum of weight of edges incident to node i in C
                ki_in = np.sum(affinityMatrix[i, :][C_members])
                deltaQ = ki_in/two_m - resolution * sigma_tot * ki / (two_m ** 2 / 2) # simplified formula. Derivation from (https://hal.archives-ouvertes.fr/hal-01231784/document)
                modularityList.append(deltaQ)
            modularityList.append(0) # append zero to avoid error by empty list
            maxQgain = np.max(modularityList)
            if maxQgain > 0:
                communityToJoin = communities[modularityList.index(maxQgain)]
                communityAssignment[i] = communityToJoin
        modularity_current = getModularity(affinityMatrix, communityAssignment, resolution)
        changeModularity = modularity_current - modularity_prev
        iterations += 1

    return communityAssignment

def LouvainPhase2(affinityMatrix, communityAssignment):
    # Merge each communities into supernodes
    # input: affinityMatrix or condensed graph
    # output: condensed graph and list of communities corresponding to the axis of graph
    num_nodes = affinityMatrix.shape[0]
    communities = list(set(communityAssignment))
    num_communities = len(communities)
    phase2graph = np.zeros([num_communities, num_communities])
    for i in range(num_communities):
        i_members = [x for x in range(num_nodes) if communityAssignment[x] == communities[i]]
        for j in range(num_communities):
            j_members = [x for x in range(num_nodes) if communityAssignment[x] == communities[j]]
            phase2graph[i, j] = np.sum(affinityMatrix[i_members, :][:, j_members]) * (1/2) ** (i == j)
            # if same communities, edges are double counted. If i == j, divide by two

    return phase2graph, communities


def Louvain(affinityMatrix, Q_threshold = 0.001, max_iter = 50, resolution = 1.0):
    # perform the two phases of Louvain community iteratively
    graph = affinityMatrix
    comm = list(range(affinityMatrix.shape[0]))
    # communityAssignmentRecord and commRefList are lists of lists, with same lengths.
    # communityAssignmentRecord stores assingment after phase 1
    # commRefList keeps the individual communities before assignment
    communityAssignmentRecord = []
    commRefList = []
    changeModularity = 1
    iterations = 0
    while changeModularity > Q_threshold and iterations < max_iter:
        graph, comm = LouvainPhase2(graph, comm)
        commRefList.append(list(comm))
        modularity_past = getModularity(graph, comm, resolution) # Note: we want to compare Q before and after reassignments in phase 1
        comm = LouvainPhase1(graph, comm, Q_threshold, max_iter, resolution)
        communityAssignmentRecord.append(list(comm))
        modularity_curr = getModularity(graph, comm, resolution)
        changeModularity = modularity_curr - modularity_past
        iterations += 1
#        print('changeModularity', changeModularity)

    for i in range(len(communityAssignmentRecord) - 2, 0, -1):
        changeDict = {}
        for j in range(len(communityAssignmentRecord[i])):
            newComm = communityAssignmentRecord[i][j]
            oldComm = commRefList[i][j]
            if newComm != oldComm:
                changeDict[oldComm] = newComm # Change key to value

        for key in changeDict:
            oldIndices = [x for x in range(len(communityAssignmentRecord[i - 1])) if communityAssignmentRecord[i - 1][x] == key]
            for j in oldIndices:
                communityAssignmentRecord[i - 1][j] = changeDict[key]
    return communityAssignmentRecord[0]


def rbfKernel(r, epsilon = 1.0):
    return np.exp(-(epsilon*r)**2)


def findThreshold(affinityMatrix):
    numFiles = affinityMatrix.shape[0]
    # Get all values in affinityMatrix. Sort in descending order
    vals = np.array([])
    for i in range(numFiles - 1):
        vals = np.append(vals, affinityMatrix[i, i + 1:])

    vals = np.sort(vals, axis = None)[::-1]

    # Initialize index values
    upperIndex = 0
    lowerIndex = len(vals) - 1

    while upperIndex != lowerIndex - 1: # Until the indices are consecutive, iterate
        midIndex = int(np.mean([lowerIndex, upperIndex]))
        adjacencyMatrix = affinityMatrix > vals[midIndex]
        # Will tally nodes visited from first node in graph by BFS. If not all nodes were visited, there are more than one component.
        nodesVisited = [0] + [i for i in range(numFiles) if adjacencyMatrix[0, i] == 1]
#        nodesVisited = [i for i in range(numFiles) if adjacencyMatrix[0, i] == 1]
        for i in nodesVisited:
            newNodes = [j for j in range(numFiles) if adjacencyMatrix[i, j] == 1 and j not in nodesVisited]
            nodesVisited += newNodes
        while len(newNodes) > 0:
            newNodes1list = []
            for i in newNodes:
                newNodes1 = [j for j in range(numFiles) if adjacencyMatrix[i, j] == 1 and j not in nodesVisited]
                newNodes1list += newNodes1
                nodesVisited += newNodes1

            newNodes = newNodes1list

        # If every node was visited, then we have exactly one component. Otherwise, there are disconnected graphs
        if len(nodesVisited) < numFiles:
            upperIndex = midIndex
        else:
            lowerIndex = midIndex

    threshold = vals[lowerIndex]

    return threshold


def AutoGraph(rmsdMatrix):
    N = rmsdMatrix.shape[0]
    affinityMatrix = rbfKernel(rmsdMatrix)
    affinityMatrix[range(N), range(N)] = 0 # set diagonal zero
    threshold = findThreshold(affinityMatrix)
    adjacencyMatrix = affinityMatrix > threshold
    filteredAffinityMatrix = affinityMatrix * adjacencyMatrix
    communityAssignment = Louvain(filteredAffinityMatrix, Q_threshold=0.0, max_iter=50, resolution=1.0)
    centroid_indices = centroid_autograph(N, communityAssignment, rmsdMatrix, threshold)
    return communityAssignment, centroid_indices


from io import BytesIO
from PIL import Image, ImageChops

from rdkit import Chem
from rdkit.Chem import AllChem, Draw, rdDepictor, rdMolTransforms
from rdkit.Chem.Draw import rdMolDraw2D
from rdkit.Chem.Draw import MolsMatrixToGridImage # new in RDKit 2023.09.1

# SVG optimization
from scour.scour import scourString

# https://greglandrum.github.io/rdkit-blog/posts/2023-05-26-drawing-options-explained.html

import numpy as np

def trim_png(img:Image.Image) -> Image.Image:
    """Removes white margin around molecular drawing.

    Args:
        img (Image.Image): input PIL Image object.

    Returns:
        Image.Image: output PIL Image object.
    """
    bg = Image.new(img.mode, img.size, img.getpixel((0,0)))
    diff = ImageChops.difference(img,bg)
    diff = ImageChops.add(diff, diff, 2.0, -100)
    bbox = diff.getbbox()
    
    if bbox:
        return img.crop(bbox)
    
    return img


def get_highlight_bonds(rdmol: Chem.Mol, atom_indices: list[int]) -> list[int] | None:
    """Get bond indices for bonds between atom indices.

    Args:
        rdmol (Chem.Mol): rdkit Chem.Mol object.
        atom_indices (list[int]): atom indices.

    Returns:
        list[int]: bond indices.
    """
    bond_indices = []
    for bond in rdmol.GetBonds():
        if bond.GetBeginAtomIdx() in atom_indices and bond.GetEndAtomIdx() in atom_indices:
            bond_indices.append(bond.GetIdx())
    
    if bond_indices:
        return bond_indices
    else:
        return None


def render_2D_mol(rdmol:Chem.Mol,
                  moldrawer:rdMolDraw2D, 
                  redraw: bool = False,
                  coordgen: bool = False,
                  legend: str = '', 
                  atom_index: bool = False, 
                  highlight_atoms: list[int] | None = None,
                  highlight_bonds: list[int] | None = None,
                  ) -> str:
    
    rdmol_2d = Chem.Mol(rdmol)
    
    if redraw or rdmol_2d.GetNumConformers() == 0:
        rdDepictor.SetPreferCoordGen(coordgen)
        rdmol_2d = Chem.RemoveHs(rdmol_2d)
        rdDepictor.Compute2DCoords(rdmol_2d)

    rdDepictor.StraightenDepiction(rdmol_2d)

    if (highlight_bonds is None) and (highlight_atoms is not None):
        # highlight bonds between the highlighted atoms
        highlight_bonds = get_highlight_bonds(rdmol_2d, highlight_atoms)

    draw_options = moldrawer.drawOptions()

    draw_options.addAtomIndices = atom_index
    # draw_options.setHighlightColour((0,.9,.9,.8)) # Cyan highlight
    # draw_options.addBondIndices = True
    # draw_options.noAtomLabels = True
    draw_options.atomLabelDeuteriumTritium = True  # D, T
    # draw_options.explicitMethyl = True
    draw_options.singleColourWedgeBonds = True
    draw_options.addStereoAnnotation = True
    # draw_options.fillHighlights = False
    # draw_options.highlightRadius = .4
    # draw_options.highlightBondWidthMultiplier = 12
    # draw_options.variableAtomRadius = 0.2
    # draw_options.variableBondWidthMultiplier = 40
    # draw_options.setVariableAttachmentColour((.5,.5,1))
    # draw_options.baseFontSize = 1.0 # default is 0.6
    # draw_options.annotationFontScale = 1
    # draw_options.rotate = 30 # rotation angle in degrees
    # draw_options.padding = 0.2 # default is 0.05

    # for atom in rdmol_2d.GetAtoms():
    #     for key in atom.GetPropsAsDict():
    #         atom.ClearProp(key)
    # if index: # index hides polar hydrogens
    #     for atom in rdmol_2d.GetAtoms():
    #        atom.SetProp("atomLabel", str(atom.GetIdx()))
    #     #    # atom.SetProp("atomNote", str(atom.GetIdx()))
    #     #    # atom.SetProp("molAtomMapNumber", str(atom.GetIdx()))        

    moldrawer.DrawMolecule(rdmol_2d, 
                           legend=legend, 
                           highlightAtoms=highlight_atoms, 
                           highlightBonds=highlight_bonds)
    moldrawer.FinishDrawing()
    
    return moldrawer.GetDrawingText()


def render_svg(rdmol: Chem.Mol, 
               width: int = 300, 
               height: int = 300,
               legend: str = '', 
               atom_index: bool = False, 
               highlight_atoms: list[int] | None = None,
               highlight_bonds: list[int] | None = None,
               redraw: bool = False,
               coordgen: bool = False,
               optimize: bool = True) -> str:
    """Draw 2D molecule in SVG format.

    Examples:
        For Jupyternotebook, wrap the output with SVG:

        >>> from IPython.display import SVG
        >>> SVG(libr[0].to_svg())

    Args:
        rdmol (Chem.Mol): rdkit Chem.Mol object.
        width (int, optional): width. Defaults to 300.
        height (int, optional): height. Defaults to 300.
        legend (str, optional): legend. Defaults to ''.
        atom_index (bool, optional): whether to show atom index. Defaults to False.
        highlight_atoms (list[int] | None, optional): atom(s) to highlight. Defaults to None.
        highlight_bonds (list[int] | None, optional): bond(s) to highlight. Defaults to None.
        redraw (bool, optional): whether to redraw. Defaults to False.
        coordgen (bool, optional): whether to use coordgen. Defaults to False.
        optimize (bool, optional): whether to optimize SVG string. Defaults to True.

    Returns:
        str: SVG string
    """

    svg_string = render_2D_mol(rdmol,
                                moldrawer = rdMolDraw2D.MolDraw2DSVG(width, height),
                                redraw = redraw, 
                                coordgen = coordgen,
                                legend = legend, 
                                atom_index = atom_index, 
                                highlight_atoms = highlight_atoms, 
                                highlight_bonds = highlight_bonds,
                                )

    if optimize:
        scour_options = {
            'strip_comments': True,
            'strip_ids': True,
            'shorten_ids': True,
            'compact_paths': True,
            'indent_type': 'none',
        }
        svg_string = scourString(svg_string, options=scour_options)

    return svg_string


def render_png(rdmol: Chem.Mol, 
               width: int = 300, 
               height: int = 300,
               legend: str = '', 
               atom_index: bool = False, 
               highlight_atoms: list[int] | None = None,
               highlight_bonds: list[int] | None = None,
               redraw: bool = False,
               coordgen: bool = False,
               trim: bool = True) -> Image.Image:
    """Draw 2D molecule in PNG format.

    Args:
        rdmol (Chem.Mol): rdkit Chem.Mol object.
        width (int, optional): width. Defaults to 300.
        height (int, optional): height. Defaults to 300.
        legend (str, optional): legend. Defaults to ''.
        atom_index (bool, optional): whether to show atom index. Defaults to False.
        highlight_atoms (list[int] | None, optional): atom(s) to highlight. Defaults to None.
        highlight_bonds (list[int] | None, optional): bond(s) to highlight. Defaults to None.
        redraw (bool, optional): whether to redraw. Defaults to False.
        coordgen (bool, optional): whether to use coordgen. Defaults to False.
    
    Returns:
        Image.Image: output PIL Image object.
    """

    png_string = render_2D_mol(rdmol,
                                moldrawer = rdMolDraw2D.MolDraw2DCairo(width, height),
                                redraw = redraw, 
                                coordgen = coordgen,
                                legend = legend, 
                                atom_index = atom_index, 
                                highlight_atoms = highlight_atoms, 
                                highlight_bonds = highlight_bonds,
                                )

    img = Image.open(BytesIO(png_string))

    if trim:
        img = trim_png(img)

    return img


def render_matrix_grid(rdmol: list[Chem.Mol],
                        legend: list[str] | None,
                        highlight_atoms: list[list[int]] | None = None,
                        highlight_bonds: list[list[int]] | None = None,
                        mols_per_row: int = 5,
                        width: int = 200,
                        height: int = 200,
                        atom_index: bool = False,
                        redraw: bool = False,
                        coordgen: bool = False,
                        svg: bool = True,
                        ) -> str | Image.Image:
    """Rendering a grid image from a list of molecules.

    Args:
        rdmol (list[Chem.Mol]): list of rdkit Chem.Mol objects.
        legend (list[str]): list of legends
        highlight_atoms (list[list[int]] | None, optional): list of atom(s) to highlight. Defaults to None.
        highlight_bonds (list[list[int]] | None, optional): list of bond(s) to highlight. Defaults to None.
        mols_per_row (int, optional): molecules per row. Defaults to 5.
        width (int, optional): width. Defaults to 200.
        height (int, optional): height. Defaults to 200.
        atom_index (bool, optional): whether to show atom index. Defaults to False.
        redraw (bool, optional): whether to redraw 2D. Defaults to False.
        coordgen (bool, optional): whether to use coordgen to depict. Defaults to False.

    Returns:
        str | Image.Image: SVG string or PIL Image object.

    Reference:
        https://greglandrum.github.io/rdkit-blog/posts/2023-10-25-molsmatrixtogridimage.html
    """

    n = len(rdmol)

    if isinstance(legend, list):
        assert len(legend) == n, "number of legends and molecules must be the same"
    elif legend is None:
        legend = ['',] * n
    
    if isinstance(highlight_atoms, list):
        assert len(highlight_atoms) == n, "number of highlights and molecules must be the same"
    elif highlight_atoms is None:
        highlight_atoms = [ (), ] * n

    if isinstance(highlight_bonds, list):
        assert len(highlight_bonds) == n, "number of highlights and molecules must be the same"    
    elif highlight_bonds is None:
        highlight_bonds = [ (), ] * n

    rdmol_matrix = []
    legend_matrix = []
    highlight_atoms_matrix = []
    highlight_bonds_matrix = []

    for i in range(0, n, mols_per_row):
        rdmol_matrix.append(rdmol[i:(i + mols_per_row)])
        legend_matrix.append(legend[i:(i + mols_per_row)])
        highlight_atoms_matrix.append(highlight_atoms[i:(i + mols_per_row)])
        highlight_bonds_matrix.append(highlight_bonds[i:(i + mols_per_row)])

    return MolsMatrixToGridImage(
            molsMatrix = rdmol_matrix,
            subImgSize = (width, height),
            legendsMatrix = legend_matrix,
            highlightAtomListsMatrix = highlight_atoms_matrix,
            highlightBondListsMatrix = highlight_bonds_matrix,
            useSVG = svg,
            returnPNG = False # whether to return PNG data (True) or a PIL object (False)
            )
    

def rescale(rdmol:Chem.Mol, factor:float=1.5) -> Chem.Mol:
    """Returns a copy of `rdmol` by a `factor`.

    Args:
        rdmol (Chem.Mol): input molecule.
        factor (float): scaling factor.
    
    Returns:
        Chem.Mol: a copy of rescaled rdkit.Chem.Mol object.
    """
    transformed_rdmol = Chem.Mol(rdmol)
    center = AllChem.ComputeCentroid(transformed_rdmol.GetConformer())
    tf = np.identity(4, np.float)
    tf[0][3] -= center[0]
    tf[1][3] -= center[1]
    tf[0][0] = tf[1][1] = tf[2][2] = factor
    AllChem.TransformMol(transformed_rdmol, tf)
    return transformed_rdmol


def rotation_matrix(axis:str, degree:float) -> np.ndarray:
    """Returns a numpy rotation matrix of shape (4,4).
    
    Args:
        axis (str): 'x' or 'y' or 'z'.
        degree (float): degree of rotation.

    Returns:
        np.ndarray: a numpy array of shape (4,4).
    """
    rad = (np.pi/180.0) * degree
    c = np.cos(rad)
    s = np.sin(rad)
    if axis.lower() == 'x':
        return np.array([
            [1., 0., 0., 0.],
            [0., c, -s,  0.],
            [0., s,  c,  0.],
            [0., 0., 0., 1.],
            ])
    elif axis.lower() == 'y':
        return np.array([
            [ c,  0., s,  0.],
            [ 0., 1., 0., 0.],
            [-s,  0., c,  0.],
            [ 0., 0., 0., 1.],
            ])
    elif axis.lower() == 'z':
        return np.array([
            [c, -s,  0., 0.],
            [s,  c,  0., 0.],
            [0., 0., 1., 0.],
            [0., 0., 0., 1.],
            ])


def rotate(rdmol:Chem.Mol, axis:str, degree:float) -> None:
    """Rotate `rdmol` around given axis and degree.

    Input `rdmol` will be modified.

    Args:
        rdmol (Chem.Mol): input molecule.
        axis (str): axis of rotation, 'x' or 'y' or 'z'.
        degree (float): degree of rotation.
    """
    try:
        conf = rdmol.GetConformer()
    except:
        AllChem.Compute2DCoords(rdmol)
        conf = rdmol.GetConformer()
    R = rotation_matrix(axis, degree)
    rdMolTransforms.TransformConformer(conf, R)



import rdworks
import rdworks.autograph
import numpy as np
from rdworks.utils import recursive_round
from rdworks.utils import convert_triu_to_symm
from rdworks.utils import serialize, deserialize

def test_serialize():
    original_data = {
        "users": ["alice", "bob", "charlie"],
        "count": 42,
        "active": True,
        "metadata": {"version": "1.0", "timestamp": "2025-10-10"}
    }
    serialized = serialize(original_data)
    deserialized = deserialize(serialized)
    assert deserialized == original_data


def test_recursive_round():

    data1 = {
        "name": "Test Data",
        "version": 1.0,
        "values": [1.23456, 2.34567, {"nested_value": 3.456789}],
        "details": {
            "temperature": 25.1234567,
            "pressure": 1013.256789,
            "measurements": [0.000123, 123.45000, 987.654321]
        },
        "another_list": [
            {"a": 1.111, "b": 2.2222},
            [3.33333, 4.444444]
        ],
        "integer_val": 10,
        "string_val": "hello"
    }

    data2 = [
        10.123,
        "string",
        [1.0, 2.3456, {"key": 9.87654321}],
        {"val1": 7.777777, "val2": [0.1, 0.02, 0.003]}
    ]

    data3 = 123.456789

    data4 = "Just a string"

    data5 = [1, 2, 3] # No floats

    print("Original data1:", data1)
    
    modified_data1_dp2 = recursive_round(data1, 2)
    
    assert modified_data1_dp2 == {'name': 'Test Data', 'version': 1.0, 'values': [1.23, 2.35, {'nested_value': 3.46}], 'details': {'temperature': 25.12, 'pressure': 1013.26, 'measurements': [0.0, 123.45, 987.65]}, 'another_list': [{'a': 1.11, 'b': 2.22}, [3.33, 4.44]], 'integer_val': 10, 'string_val': 'hello'}
    print("\nModified data1 (2 decimal places):", modified_data1_dp2)
    
    modified_data1_dp0 = recursive_round(data1, 0)
    assert modified_data1_dp0 == {'name': 'Test Data', 'version': 1.0, 'values': [1.0, 2.0, {'nested_value': 3.0}], 'details': {'temperature': 25.0, 'pressure': 1013.0, 'measurements': [0.0, 123.0, 988.0]}, 'another_list': [{'a': 1.0, 'b': 2.0}, [3.0, 4.0]], 'integer_val': 10, 'string_val': 'hello'}
    print("\nModified data1 (0 decimal places):", modified_data1_dp0)

    print("\nOriginal data2:", data2)
    modified_data2_dp3 = recursive_round(data2, 3)
    assert modified_data2_dp3 == [10.123, 'string', [1.0, 2.346, {'key': 9.877}], {'val1': 7.778, 'val2': [0.1, 0.02, 0.003]}]
    print("\nModified data2 (3 decimal places):", modified_data2_dp3)

    print("\nOriginal data3:", data3)
    modified_data3_dp1 = recursive_round(data3, 1)
    assert modified_data3_dp1 == 123.5
    print("\nModified data3 (1 decimal place):", modified_data3_dp1)

    print("\nOriginal data4:", data4)
    modified_data4_dp2 = recursive_round(data4, 2)
    assert modified_data4_dp2 == 'Just a string'
    print("\nModified data4 (2 decimal places):", modified_data4_dp2) # Should be unchanged

    print("\nOriginal data5:", data5)
    modified_data5_dp2 = recursive_round(data5, 2)
    assert modified_data5_dp2 == [1, 2, 3]
    print("\nModified data5 (2 decimal places):", modified_data5_dp2) # Should be unchanged

    # Example of invalid decimal_places
    try:
        recursive_round(data1, -1)
    except ValueError as e:
        print(f"\nError caught: {e}")

    try:
        recursive_round(data1, 1.5)
    except ValueError as e:
        print(f"Error caught: {e}")


def test_autograph():
    N = 50
    upper_triangle_values = 5.0 *np.random.rand(N*(N-1)//2)
    rmsdMatrix = convert_triu_to_symm(upper_triangle_values)
    com, cen = rdworks.autograph.NMRCLUST(rmsdMatrix)
    assert len(com) == N
    assert len(set(com)) == len(cen)
    
    com, cen = rdworks.autograph.DynamicTreeCut(rmsdMatrix)
    assert len(com) == N
    assert len(set(com)) == len(cen)
    
    com, cen = rdworks.autograph.RCKmeans(rmsdMatrix)
    assert len(com) == N
    assert len(set(com)) == len(cen)

    com, cen  = rdworks.autograph.AutoGraph(rmsdMatrix)
    assert len(com) == N
    assert len(set(com)) == len(cen)


# 2020-05-22

# Kiyoto Aramis Tanemura

# The Ward algorithm for conformational clustering is used in building Markov State Model (DOI: 10.1021/acs.jctc.6b01238). To circumvent threshold selection, we will apply the dynamic tree cut method, used in conjunction to the Ward dendogram for automated conformational clustering (DOI: 10.1186/s13321-017-0208-0).

import numpy as np

from scipy.cluster.hierarchy import linkage, to_tree
from scipy.spatial.distance import squareform

from .centroid import centroid_medoid


def get_ward_dendrogram(rmsd_mat):
    '''Use Scipy functions to obtain dendrogram using Ward method
    Returns the dendrogram (refer to Scipy linkage output format) as a ClusterNode object'''
    dend = linkage(squareform(rmsd_mat), method = 'ward', optimal_ordering = True)
    return dend


def goLeftmost(node, path, path_record):
    '''provided a node, travel left until a leaf is reached'''
    curr_node = node
    if path in path_record:
        return path
    while not curr_node.is_leaf():
        curr_node = curr_node.get_left()
        path += ('l')
    return path


def travelDown(ref_node, rel_path):
    '''reach a node below, provided a starting node an path'''
    curr_node = ref_node
    for i in rel_path:
        if i == 'l':
            curr_node = curr_node.get_left()
        elif i == 'r':
            curr_node = curr_node.get_right()
    return curr_node


def getHeights(root):
    '''return heights of nonleaf nodes and their corresponding paths'''
    heights = []
    path_recorded = []
    path = '.'
    curr_node = root
    if root.is_leaf():
        return [0], [path]
    while 'l' in list(path) or not curr_node.is_leaf():
        path = goLeftmost(curr_node, path, path_recorded)
        path = path[:-1]
        curr_node = travelDown(root, path)
        if path not in path_recorded:
            heights.append(curr_node.dist)
            path_recorded.append(path)
            path += 'r'
            curr_node = travelDown(root, path)
    return heights, path_recorded


def treeCutCore(H, I, tau = 5):
    '''Determine significant clusters provided one calibration value'''
    H_hat = H - I
    trans_indices = [x for x in range(len(H)-1) if H_hat[x] > 0 and H_hat[x+1] < 0]
    breakpoints = []
    for i in trans_indices:
        back_index = 1
        while H_hat[i-back_index] > 0 and back_index <= i:
            back_index += 1
        breakpoints.append(i - back_index + 1)
    # Find significant breakpoints
    significant_breakpoints = [breakpoints[x] for x in range(len(breakpoints)) if trans_indices[x] - breakpoints[x] > tau]
    return significant_breakpoints


def adaptiveTreecutCore(H, tau = 5):
    '''Perform treeCutCore at mean height. If no significant breakpoints are detected,
    continue the operation below and above the mean.'''
    if len(H) == 0:
        return []
    lm = np.mean(H)
    lu = np.mean([lm, np.max(H)])
    ld = np.mean([lm, np.min(H)])
    bps = treeCutCore(H, lm, tau)
    if len(bps) == 0:
        bps = treeCutCore(H, ld, tau)
        if len(bps) == 0:
            bps = treeCutCore(H, lu, tau)
    return bps


def getClusterNodeIndices(comprehensive_path_list, cluster_substring):
    # Given a root string, return all permutations as indices
    return [x for x in range(len(comprehensive_path_list)) if comprehensive_path_list[x].startswith(cluster_substring)]


# After looking at the Java implementation, I suspect the breakpoints, corresponding to the indices of heights (distance value of nonleaf nodes),
# also correspond to the indices of leaves. Try clustering only by collecting all breakpoints, then using the indices to subset leaves.
def dynamicTreeCut(tree, n, tau = 5):
    allHeights, allPaths = getHeights(tree)
    allHeights = [0] + allHeights + [0] # Sandwitch the heights with zero so that the ends can be included or left out from major clusters.
    breakpoints = [0,-1]
    updateList = [-1]
    while len(updateList) > 0:
        updateList = []
        for i in range(len(breakpoints) - 1):
            Hi = allHeights[breakpoints[i]:breakpoints[i+1]]
            cutpoints = adaptiveTreecutCore(Hi, tau)
            updateList += [x for x in cutpoints if x not in breakpoints]
        breakpoints += updateList
    return breakpoints


def report_assingments(breakpoints, tree):
    '''To standardize outputs with the other clustering algorithms, this function takes the list of ClusterNodes produced in dynamicTreeCut and returns a list of ints specifying the cluster assignment'''
    leaves = tree.pre_order(lambda x: x.id)
    n = len(leaves)
    comm_assing = np.zeros(n, int)
    for i in range(len(breakpoints) - 1):
        members = leaves[breakpoints[i]:breakpoints[i+1]]
        comm_assing[members] = i
    return comm_assing.tolist()


def DynamicTreeCut(rmsdMatrix, tau = 5): # Ward clustering
    N = rmsdMatrix.shape[0]
    dend = get_ward_dendrogram(rmsdMatrix)
    tree = to_tree(dend)
    breakpoints = dynamicTreeCut(tree, N, tau)
    communityAssignment = report_assingments(breakpoints, tree)
    centroid_indices = centroid_medoid(communityAssignment, rmsdMatrix)
    return communityAssignment, centroid_indices


import pandas as pd
import numpy as np


def centroid_medoid(communityAssignment, rmsdMatrix) -> list:
    """returns a list of centroids based on medoids

    Medoids are representative objects of a data set or a cluster within a data set 
    whose sum of dissimilarities to all the objects in the cluster is minimal.

    Args:
        communityAssignment: (list) list of community assignment correspoinding to the index of fileList
        rmsdMatrix: (numpy array) Matrix containing pairwise atomic RMSD between all conformers
    
    Returns:
        a list of centroids
    """
    N = rmsdMatrix.shape[0]
    community_indices = list(set(communityAssignment))
    centroids = []
    for C in community_indices:
        C_members = [x for x in range(N) if communityAssignment[x] == C]
        community_submatrix = rmsdMatrix[C_members, :][:, C_members]
        dist_sum = np.sum(community_submatrix, axis = 1)
        centroids.append(C_members[np.argmin(dist_sum)])
    return centroids


def diekstra(filtered_rmsd_matrix_community, i):
    '''Use Dijkstra's algorithm to find shortest path from index i to all other nodes'''
    # initialize lists
    visited = []
    unvisited = [x for x in range(filtered_rmsd_matrix_community.shape[0])]
    record = [np.inf for x in range(filtered_rmsd_matrix_community.shape[0])]#{x: np.inf for x in range(graph.shape[0])}
    record[i] = 0
    lastNode = [-1 for x in record]
    # repeat until all nodes have been visited
    while len(unvisited) > 0:
        visit_index = unvisited[np.argmin([record[x] for x in unvisited])]
        unvisited_neighbors = [x for x in unvisited if filtered_rmsd_matrix_community[visit_index, x] > 0]
        # Calculate distance to unvisited neighbor. If value is shorter than recorded, update distance.
        updateDist = filtered_rmsd_matrix_community[visit_index, :] + record[visit_index]
        for j in unvisited_neighbors:
            record[j] = np.min([updateDist[j], record[j]])
            if updateDist[j] < record[j]:
                lastNode[j] = visit_index
        # update visited/unvisited node list
        unvisited.remove(visit_index)
        visited.append(visit_index)
    return record, lastNode


def centroid_betweenness(num, communityAssignment, filtered_rmsd_matrix):
    # Provided with list of conformers assigned to communities, choose representative centroid by conformers of maximum in community betweenness
    # inputs
    # fileList: (list) names of xyz files for each conformer
    # communityAssignment: (list) list of community assignment correspoinding to the index of fileList
    # filtered_rmsd_matrix: {np.array) RMSD matrix between conformers, except assigning distances above threshold to zero
    communityList = list(set(communityAssignment))
    centralNodes = []
    comm_size = []
    for C in communityList:
        C_members = [x for x in range(num) if communityAssignment[x] == C]
        C_member_files = [x for x in C_members]
        comm_size.append(len(C_members))
        community_subgraph = filtered_rmsd_matrix[C_members, :][:, C_members]
        community_betweenness = np.zeros(len(C_members))
        for i in range(len(C_member_files)):
            record, lastnode = diekstra(community_subgraph, i)
            for j in range(len(C_member_files)):
                previous_node = lastnode[j]
                while previous_node != -1:
                    community_betweenness[previous_node] += 1
                    previous_node = lastnode[previous_node]
        max_betweenness_index = np.argmax(community_betweenness)
        centralNodes.append(C_member_files[max_betweenness_index])

    # Sort centers by size of clusters in descending order
    centralDf = pd.DataFrame({'size': comm_size}, index = centralNodes)
    centralDf.sort_values(by = 'size', ascending = False, inplace = True)

    return list(centralDf.index)


def centroid_autograph(N, communityAssignment, rmsdMatrix, threshold, centroid_selection='betweenness', filteredAffinityMatrix=None):
    '''Return file names of conformers designated as centroids. If energy is provided, find lowest energy conformers in each cluster. Otherwise choose by maximum in-cluster weighted degree'''
    if centroid_selection == 'betweenness':
        return centroid_betweenness(N, communityAssignment, rmsdMatrix * rmsdMatrix < np.sqrt(-np.log(threshold))) # add filtered rmsd matrix
    else:
        print('centroid criterion not recognized. Use keywords "degree", "eccentricity", or "betweenness" for centroid_selection or provide an energy output to base the selection')


from pathlib import Path
from rdkit import Chem

import rdworks
from rdworks import Mol, MolLibr

datadir = Path(__file__).parent.resolve() / "data"


# python >=3.12 raises SyntaxWarning: invalid escape sequence
# To address this warning in general, we can make the string literal a raw string literal r"...". 
# Raw string literals do not process escape sequences. 
# For example, r"\n" is treated simply as the characters \ and n and not as a newline escape sequence.
drug_smiles = [
    "Fc1cc(c(F)cc1F)C[C@@H](N)CC(=O)N3Cc2nnc(n2CC3)C(F)(F)F", # [0]
    r"O=C(O[C@@H]1[C@H]3C(=C/[C@H](C)C1)\C=C/[C@@H]([C@@H]3CC[C@H]2OC(=O)C[C@H](O)C2)C)C(C)(C)CC",
    "C[C@@H](C(OC(C)C)=O)N[P@](OC[C@@H]1[C@H]([C@@](F)([C@@H](O1)N2C=CC(NC2=O)=O)C)O)(OC3=CC=CC=C3)=O",
    "C1CNC[C@H]([C@@H]1C2=CC=C(C=C2)F)COC3=CC4=C(C=C3)OCO4",
    "CC1=C(C=NO1)C(=O)NC2=CC=C(C=C2)C(F)(F)F",
    "CN1[C@@H]2CCC[C@H]1CC(C2)NC(=O)C3=NN(C4=CC=CC=C43)C", # [5] - Granisetron
    "CCCN1C[C@@H](C[C@H]2[C@H]1CC3=CNC4=CC=CC2=C34)CSC",
    "CCC1=C(NC2=C1C(=O)C(CC2)CN3CCOCC3)C", # [7] Molidone
    r"C[C@H]1/C=C/C=C(\C(=O)NC2=C(C(=C3C(=C2O)C(=C(C4=C3C(=O)[C@](O4)(O/C=C/[C@@H]([C@H]([C@H]([C@@H]([C@@H]([C@@H]([C@H]1O)C)O)C)OC(=O)C)C)OC)C)C)O)O)/C=N/N5CCN(CC5)C)/C",
    r"C=CC1=C(N2[C@@H]([C@@H](C2=O)NC(=O)/C(=N\O)/C3=CSC(=N3)N)SC1)C(=O)O",
    "CC1=C(N=CN1)CSCCNC(=NC)NC#N", # [10] - Cimetidine
    """C1=C(N=C(S1)N=C(N)N)CSCC/C(=N/S(=O)(=O)N)/N""",
    "C1CC(CCC1C2=CC=C(C=C2)Cl)C3=C(C4=CC=CC=C4C(=O)C3=O)O",
    "CN(CC/C=C1C2=CC=CC=C2SC3=C/1C=C(Cl)C=C3)C",
    "CN(C)CCCN1C2=CC=CC=C2CCC3=C1C=C(C=C3)Cl",
    "CN1CCCC(C1)CC2C3=CC=CC=C3SC4=CC=CC=C24", # [15] - Methixene
    "CCN(CC)C(C)CN1C2=CC=CC=C2SC3=CC=CC=C31",
    "CC(=O)OC1=CC=CC=C1C(=O)O",
    "C1=CC(=C(C=C1F)F)C(CN2C=NC=N2)(CN3C=NC=N3)O",
    "CC(=O)NC[C@H]1CN(C(=O)O1)C2=CC(=C(C=C2)N3CCOCC3)F", # [19]
    ]

drug_names = [
    "Sitagliptin", "Simvastatin", "Sofosbuvir", "Paroxetine", "Leflunomide",
    "Granisetron", "Pergolide", "Molindone", "Rifampin", "Cefdinir",
    "Cimetidine", "Famotidine", "Atovaquone", "Chlorprothixene", "Clomipramine",
    "Methixene",  "Ethopropazine", "Aspirin", "Fluconazole", "Linezolid",
    ]


def test_init_mol():
    mol = Mol(drug_smiles[0], drug_names[0])
    assert mol.num_confs == 0
    assert mol.name == drug_names[0]
    rdmol = Chem.MolFromSmiles(drug_smiles[0])
    rdmol.SetProp('_Name', drug_names[0])
    mol = Mol(rdmol, drug_names[0])
    assert mol.rdmol.GetProp('_Name') == drug_names[0]
    assert mol.name == drug_names[0]


def test_has_substr():
    mol = Mol('c1cc(C(=O)O)c(OC(=O)C)cc1')
    assert mol.has_substr('OC(=O)C')


def test_remove_stereo():
    m = Mol("C/C=C/C=C\\C", "double_bond")
    assert m.remove_stereo().smiles == "CC=CC=CC"


def test_complete_stereoisomers():
    m = Mol("CC=CC", "double_bond")
    assert m.is_stereo_specified is False, "double bond stereo is not properly handled"
    libr = rdworks.complete_stereoisomers(m)
    assert libr.count() == 2, "cis and trans are expected"
    assert all([_.is_stereo_specified for _ in libr])
    expected_canonical_smiles = [r'C/C=C/C', r'C/C=C\C']
    canonical_smiles = [_.smiles for _ in libr]
    difference = set(expected_canonical_smiles) - set(canonical_smiles)
    assert len(difference) == 0

    # 0 out of 3 atom stereocenters is specified
    m = Mol("N=C1OC(CN2CC(C)OC(C)C2)CN1", "stereoisomer")
    assert m.is_stereo_specified is False
    libr = rdworks.complete_stereoisomers(m)
    assert libr.count() == 6, "0 out of 3 atom stereocenters is specified"
    assert all([_.is_stereo_specified for _ in libr])
    expected_names = [f'stereoisomer.{i}' for i in [1,2,3,4,5,6]]
    names = [_.name for _ in libr]
    assert names == expected_names
    expected_canonical_smiles = [
        'C[C@@H]1CN(C[C@@H]2CNC(=N)O2)C[C@H](C)O1',
        'C[C@@H]1CN(C[C@H]2CNC(=N)O2)C[C@H](C)O1',
        'C[C@@H]1CN(C[C@@H]2CNC(=N)O2)C[C@@H](C)O1',
        'C[C@@H]1CN(C[C@H]2CNC(=N)O2)C[C@@H](C)O1',
        'C[C@H]1CN(C[C@@H]2CNC(=N)O2)C[C@H](C)O1',
        'C[C@H]1CN(C[C@H]2CNC(=N)O2)C[C@H](C)O1',
        ]
    canonical_smiles = [_.smiles for _ in libr]
    difference = set(expected_canonical_smiles) - set(canonical_smiles)
    assert len(difference) == 0
    
    # 1 out of 3 atom stereocenters is specified
    m = Mol("N=C1OC(CN1)CN2CC(O[C@H](C2)C)C", "stereoisomer") 
    assert m.is_stereo_specified is False
    libr = rdworks.complete_stereoisomers(m)
    assert libr.count() == 4, "1 out of 3 atom stereocenters is specified"
    assert all([_.is_stereo_specified for _ in libr])
    expected_names = [f'stereoisomer.{i}' for i in [1,2,3,4]]
    names = [_.name for _ in libr]
    assert names == expected_names
    expected_canonical_smiles = [
        'C[C@@H]1CN(C[C@@H]2CNC(=N)O2)C[C@H](C)O1',
        'C[C@@H]1CN(C[C@H]2CNC(=N)O2)C[C@H](C)O1',
        'C[C@H]1CN(C[C@@H]2CNC(=N)O2)C[C@H](C)O1',
        'C[C@H]1CN(C[C@H]2CNC(=N)O2)C[C@H](C)O1',
        ]
    canonical_smiles = [_.smiles for _ in libr]
    difference = set(expected_canonical_smiles) - set(canonical_smiles)
    assert len(difference) == 0

    # 2 out of 3 atom stereocenters are specified
    m = Mol("N=C1OC(CN1)CN2C[C@H](O[C@H](C2)C)C", "stereoisomer") 
    assert m.is_stereo_specified is False
    libr = rdworks.complete_stereoisomers(m)
    assert libr.count() == 2, "2 out of 3 atom stereocenters is specified"
    assert all([_.is_stereo_specified for _ in libr])
    expected_names = [f'stereoisomer.{i}' for i in [1,2]]
    names = [_.name for _ in libr]
    assert names == expected_names
    expected_canonical_smiles = ['C[C@@H]1CN(C[C@@H]2CNC(=N)O2)C[C@H](C)O1',
                                    'C[C@@H]1CN(C[C@H]2CNC(=N)O2)C[C@H](C)O1',
                                    ]
    canonical_smiles = [_.smiles for _ in libr]
    difference = set(expected_canonical_smiles) - set(canonical_smiles)
    assert len(difference) == 0
    
    # 3 out of 3 atom stereocenters are specified
    m = Mol("N=C1O[C@@H](CN1)CN2C[C@H](O[C@H](C2)C)C", "stereoisomer") 
    assert m.is_stereo_specified is True
    libr = rdworks.complete_stereoisomers(m)
    assert libr.count() == 1, "3 out of 3 atom stereocenters is specified"
    assert all([_.is_stereo_specified for _ in libr])
    expected_names = [f'stereoisomer']
    names = [_.name for _ in libr]
    assert names == expected_names
    expected_canonical_smiles = ['C[C@@H]1CN(C[C@@H]2CNC(=N)O2)C[C@H](C)O1']
    canonical_smiles = [_.smiles for _ in libr]
    difference = set(expected_canonical_smiles) - set(canonical_smiles)
    assert len(difference) == 0

    # for 20 molecules
    isomer_libr = MolLibr()
    for mol in MolLibr(drug_smiles, drug_names):
        isomer_libr += rdworks.complete_stereoisomers(mol)
    assert isomer_libr.count() >= 25


def test_num_stereoisomers():
    """count all possible stereoisomers ignoring current stereochemistry assignment"""
    m = Mol('Cc1nc2c(-c3ccc(Cl)cc3F)nc(N3CCN(C(=O)C(F)F)CC3)cn2c(=O)c1C')
    assert m.num_stereoisomers == 1
    m = Mol('CN1C=C([C@H]2CN(C3=NC(C4=CC=C(Cl)C=C4F)=C4N=C5CCCCN5C(=O)C4=C3)CCO2)C=N1')
    assert m.num_stereoisomers == 2
    m = Mol('Cc1nc2c(-c3ccc(Cl)cc3F)nc(N3CCN(S(=O)(=O)N4C[C@@H](C)O[C@@H](C)C4)CC3)cn2c(=O)c1C')
    assert m.num_stereoisomers == 3
    m = Mol('Cc1cc([C@H]2CN(c3cc4nc(C)c(C)c(=O)n4c(-c4ccc(Cl)cc4F)n3)C[C@@H](C)O2)ccn1')
    assert m.num_stereoisomers == 4


def test_complete_tautomers():
    m = Mol("Oc1c(cccc3)c3nc2ccncc12", "tautomer")
    libr = rdworks.complete_tautomers(m)
    assert libr.count() == 3
    expected_names = ['tautomer.1','tautomer.2','tautomer.3']
    names = [_.name for _ in libr]
    assert names == expected_names
    expected_canonical_smiles = [
        'O=c1c2c[nH]ccc-2nc2ccccc12',
        'O=c1c2ccccc2[nH]c2ccncc12',
        'Oc1c2ccccc2nc2ccncc12',
        ]
    canonical_smiles = [_.smiles for _ in libr]
    difference = set(expected_canonical_smiles) - set(canonical_smiles)
    assert len(difference) == 0
    
    m = Mol("CCCCNC(=O)[C@@H]1CCCN1[C@@H](CC)c1nnc(Cc2ccc(C)cc2)o1")
    libr = rdworks.complete_tautomers(m)
    assert libr.count() == 5


# 2020-05-20

# Kiyoto Aramis Tanemura

# We consider the use of representative conformation K-means to benchmark against AutoGraph as a method which does not require specification of number of clusters or threshold.
# Original algorithm found at DOI: 10.1186/s13321-017-0208-0

import numpy as np
from random import sample
from math import factorial

def kmedoid(rmsdMatrix, k = 2):
    # Generic k-medoid function with rmsd matrix as input
    n = rmsdMatrix.shape[0]
    medoids = sample(range(n), k)
    prev_medoids = []
    classification = np.zeros(n, dtype = int).tolist()
    while medoids != prev_medoids:
        for i in range(n):
            prev_medoids = medoids
            min_index = np.argmin(rmsdMatrix[i,:][medoids])
            classification[i] = min_index
        for j in range(k):
            members = [x for x in range(n) if classification[x] == j]
            sub_rmsd = rmsdMatrix[members,:][:, members]
            center_index = np.argmin(np.sum(sub_rmsd, axis = 0))
            medoids[j] = center_index

    return classification, medoids


def comb(n, r):
    if n < r:
        return 1
    return factorial(n) / (factorial(r) * factorial(n-r))


def MSQb(rmsdMatrix, medoids):
    sub_rmsd = rmsdMatrix[medoids, :][:, medoids]
    return np.sum(sub_rmsd) / (2 * comb(len(medoids), 2))


def MSQw(rmsdMatrix,classification):
    tally = 0
    n = len(classification)
    for i in set(classification):
        members = [x for x in range(n) if classification[x] == i]
        sub_rmsd = rmsdMatrix[members, :][:, members]
        tally += np.sum(sub_rmsd) / (2 * comb(len(members), 2))
    return tally / len(set(classification))


def SMA(MSQb_list, W = 10):
    if len(MSQb_list) >= W:
        return np.mean(MSQb_list[-10:])
    return -1


def RCKmeans(rmsdMatrix):
    m = rmsdMatrix.shape[0]
    K_MSQb = [0, 0]
    prevSMA = -1
    for k in range(2,m):
        MSQw_list = np.zeros(100, dtype = int).tolist()
        MSQb_list = np.zeros(100, dtype = int).tolist()
        for i in range(100):
            classification, medoids = kmedoid(rmsdMatrix, k)
            MSQw_list[i] = MSQw(rmsdMatrix, classification)
            MSQb_list[i] = MSQb(rmsdMatrix, medoids)
        K_MSQb.append(MSQb_list[np.argmin(MSQw_list)])
        currSMA = SMA(K_MSQb, 10)
        if currSMA < prevSMA:
            return kmedoid(rmsdMatrix, np.argmax(K_MSQb))
        prevSMA = currSMA


import os
import resource
import subprocess
import json
import tempfile
import logging
import shutil
import re

from pathlib import Path
from types import SimpleNamespace

from rdkit import Chem
from rdkit.Chem import rdmolops
from rdkit.Geometry import Point3D


logger = logging.getLogger(__name__)


# In ASE, the default energy unit is eV (electron volt).
# It will be converted to kcal/mol
# CODATA 2018 energy conversion factor
hartree2ev = 27.211386245988
hartree2kcalpermol = 627.50947337481
ev2kcalpermol = 23.060547830619026


class GFN2xTB:
    def __init__(self, molecule: Chem.Mol, ncores: int | None = None):
        assert isinstance(molecule, Chem.Mol), "molecule is not rdkit.Chem.Mol type"
        assert molecule.GetConformer().Is3D(), "molecule is not a 3D conformer"
        assert self.is_xtb_ready(), "xtb is not accessible"

        self.rdmol = molecule
        self.charge = rdmolops.GetFormalCharge(self.rdmol)
        self.natoms = molecule.GetNumAtoms()
        self.symbols = [ atom.GetSymbol() for atom in molecule.GetAtoms() ]
        self.positions = molecule.GetConformer().GetPositions().tolist()

        if ncores is None:
            ncores = os.cpu_count()

        # Parallelisation
        os.environ['OMP_STACKSIZE'] = '4G'
        os.environ['OMP_NUM_THREADS'] = f'{ncores},1'
        os.environ['OMP_MAX_ACTIVE_LEVELS'] = '1'
        os.environ['MKL_NUM_THREADS'] = f'{ncores}'
        
        # unlimit the system stack
        resource.setrlimit(resource.RLIMIT_STACK, (resource.RLIM_INFINITY, resource.RLIM_INFINITY))


    @staticmethod
    def is_xtb_ready() -> bool:
        """Check if xtb is available.

        Returns:
            bool: True if `xtb` is available, False otherwise.
        """
        return shutil.which('xtb') is not None


    @staticmethod
    def is_optimize_ready() -> bool:
        try:
            h2o = [
                '$coord',
                ' 0.00000000000000      0.00000000000000     -0.73578586109551      o',
                ' 1.44183152868459      0.00000000000000      0.36789293054775      h',
                '-1.44183152868459      0.00000000000000      0.36789293054775      h',
                '$end',
            ]

            with tempfile.TemporaryDirectory() as temp_dir:
                test_geometry = os.path.join(temp_dir, 'coord')
                with open(test_geometry, 'w') as f:
                    f.write('\n'.join(h2o))
                proc = subprocess.run(['xtb', test_geometry, '--opt'],
                                      cwd=temp_dir,
                                      capture_output=True, 
                                      text=True,
                                      encoding='utf-8')
                assert proc.returncode == 0

            return True

        except:
            print("""                          
Conda installed xTB has the Fortran runtime error in geometry optimization. 
Please install xtb using the compiled binary:
                    
$ wget https://github.com/grimme-lab/xtb/releases/download/v6.7.1/xtb-6.7.1-linux-x86_64.tar.xz
$ tar -xf xtb-6.7.1-linux-x86_64.tar.xz
$ cp -r xtb-dist/bin/*      /usr/local/bin/
$ cp -r xtb-dist/lib/*      /usr/local/lib/
$ cp -r xtb-dist/include/*  /usr/local/include/
$ cp -r xtb-dist/share      /usr/local/ """)
            
            return False


    @staticmethod
    def is_cpx_ready() -> bool:
        """Checks if the CPCM-X command-line tool, `cpx`, is accessible in the system.

        Returns:
            bool: True if the cpx is found, False otherwise.
        """
        return shutil.which('cpx') is not None
    

    @staticmethod
    def is_cpcmx_ready() -> bool:
        """Checks if xtb works with the `--cpcmx` option.

        xtb distributed by the conda does not include CPCM-X function (as of June 17, 2025). 
        xtb installed from the github source codes by using meson and ninja includes it.

        Returns:
            bool: True if the --cpcmx option is working, False otherwise.
        """
        if GFN2xTB.is_xtb_ready():
            with tempfile.TemporaryDirectory() as temp_dir: # tmpdir is a string
                cmd = ['xtb', '--cpcmx']
                proc = subprocess.run(cmd, cwd=temp_dir, capture_output=True, text=True, encoding='utf-8')
                # we are expecting an error because no input file is given
                assert proc.returncode != 0
                for line in proc.stdout.split('\n'):
                    line = line.strip()
                    if 'CPCM-X library was not included' in line:
                        return False
        
        return True


    @staticmethod
    def is_ready() -> bool:
        """Check if `xtb` and `cpx` are accessible and `xtb --cpcmx` are available.

        Returns:
            bool: True if both `xtb` and `cpx` are accessible, False otherwise.
        """
        return all([GFN2xTB.is_xtb_ready(),
                    GFN2xTB.is_cpx_ready(),
                    GFN2xTB.is_cpcmx_ready(),
                    GFN2xTB.is_optimize_ready()])
    

    @staticmethod
    def version() -> str | None:
        """Check xtb version.

        Returns:
            str | None: version statement.
        """
        if GFN2xTB.is_xtb_ready():
            with tempfile.TemporaryDirectory() as temp_dir: # tmpdir is a string
                cmd = ['xtb', '--version']
                proc = subprocess.run(cmd, cwd=temp_dir, capture_output=True, text=True, encoding='utf-8')
                assert proc.returncode == 0, "GFN2xTB() Error: xtb not available"
                match = re.search('xtb\s+version\s+(?P<version>[\d.]+)', proc.stdout)
                if match:
                    return match.group('version')
            
        return None
    

    def to_xyz(self) -> str:
        """Export to XYZ formatted string.

        Returns:
            str: XYZ formatted string
        """
        lines = [f'{self.natoms}', ' ']
        for e, (x, y, z) in zip(self.symbols, self.positions):
            lines.append(f'{e:5} {x:23.14f} {y:23.14f} {z:23.14f}')
        
        return '\n'.join(lines)


    def to_turbomole_coord(self, bohr: bool = False) -> str:
        """Returns TURBOMOLE coord file formatted strings.

        Turbomole coord file format:

            - It starts with the keyword `$coord`.
            - Each line after the $coord line specifies an atom, consisting of:
                - Three real numbers representing the Cartesian coordinates (x, y, z).
                - A string for the element name.
                - Optional: an "f" label at the end to indicate that the atom's coordinates are frozen during optimization.
            - Coordinates can be given in Bohr (default), Ångström (`$coord angs`), or fractional coordinates (`$coord frac`).
            - Optional data groups like periodicity (`$periodic`), lattice parameters (`$lattice`), and cell parameters (`$cell`) can also be included. 
            - Regarding precision:
                The precision of the coordinates is crucial for accurate calculations, especially geometry optimizations.
                Tools like the TURBOMOLEOptimizer might check for differences in atomic positions with a tolerance of 1e-13. 

        Args:
            bohr (bool): whether to use Bohr units of the coordinates. Defaults to False.
                Otherwise, Angstrom units will be used.
        
        Returns:
            str: TURBOMOLE coord formatted file.
        """
        if bohr:
            lines = ["$coord"]
        else:
            lines = ["$coord angs"]

        for (x, y, z), e in zip(self.positions, self.symbols):
            lines.append(f"{x:20.15f} {y:20.15f} {z:20.15f} {e}")
        
        lines.append("$end")

        return '\n'.join(lines)


    def load_xyz(self, geometry_input_path: Path) -> Chem.Mol:
        """Load geometry.

        Args:
            geometry_input_path (Path): pathlib.Path to the xyz 

        Returns:
            Chem.Mol: rdkit Chem.Mol object.
        """
        rdmol_opt = Chem.Mol(self.rdmol)
        with open(geometry_input_path, 'r') as f:
            for lineno, line in enumerate(f):
                if lineno == 0:
                    assert int(line.strip()) == self.natoms
                    continue
                elif lineno == 1: # comment or title
                    continue
                (symbol, x, y, z) = line.strip().split()
                x, y, z = float(x), float(y), float(z)
                atom = rdmol_opt.GetAtomWithIdx(lineno-2)
                assert symbol == atom.GetSymbol()
                rdmol_opt.GetConformer().SetAtomPosition(atom.GetIdx(), Point3D(x, y, z))
        
        return rdmol_opt


    def load_wbo(self, wbo_path: Path) -> dict[tuple[int, int], float]:
        """Load Wiberg bond order.

        singlepoint() creates a wbo output file.

        Args:
            wbo_path (Path): path to the wbo file.

        Returns:
            dict(tuple[int, int], float): { (i, j) : wbo, ... } where i and j are atom indices for a bond.
        """

        with open(wbo_path, 'r') as f:
            # Wiberg bond order (WBO)
            Wiberg_bond_orders = {}
            for line in f:
                line = line.strip()
                if line:
                    # wbo output has 1-based indices
                    (i, j, wbo) = line.split()
                    # changes to 0-based indices
                    i = int(i) - 1
                    j = int(j) - 1
                    # wbo ouput indices are ascending order
                    ij = (i, j) if i < j else (j, i)
                    Wiberg_bond_orders[ij] = float(wbo)

            return Wiberg_bond_orders


    def singlepoint(self, water: str | None = None, verbose: bool = False) -> SimpleNamespace:
        """Calculate single point energy.
        
        Total energy from xtb output in atomic units (Eh, hartree) is converted to kcal/mol.

        Options:
            ```sh
            -c, --chrg INT
                specify molecular charge as INT, overrides .CHRG file and xcontrol option
            
            --scc, --sp
                performs a single point calculation
            
            --gfn INT
                specify parametrisation of GFN-xTB (default = 2)
            
            --json
                write xtbout.json file
            
            --alpb SOLVENT [STATE]
                analytical linearized Poisson-Boltzmann (ALPB) model,
                available solvents are acetone, acetonitrile, aniline, benzaldehyde,
                benzene, ch2cl2, chcl3, cs2, dioxane, dmf, dmso, ether, ethylacetate, furane,
                hexandecane, hexane, methanol, nitromethane, octanol, woctanol, phenol, toluene,
                thf, water.
                The solvent input is not case-sensitive. The Gsolv
                reference state can be chosen as reference, bar1M, or gsolv (default).

            -g, --gbsa SOLVENT [STATE]
                generalized born (GB) model with solvent accessable surface (SASA) model,
                available solvents are acetone, acetonitrile, benzene (only GFN1-xTB), CH2Cl2,
                CHCl3, CS2, DMF (only GFN2-xTB), DMSO, ether, H2O, methanol,
                n-hexane (only GFN2-xTB), THF and toluene.
                The solvent input is not case-sensitive.
                The Gsolv reference state can be chosen as reference, bar1M, or gsolv (default).

            --cosmo SOLVENT/EPSILON
                domain decomposition conductor-like screening model (ddCOSMO),
                available solvents are all solvents that are available for alpb.
                Additionally, the dielectric constant can be set manually or an ideal conductor
                can be chosen by setting epsilon to infinity.

            --tmcosmo SOLVENT/EPSILON
                same as --cosmo, but uses TM convention for writing the .cosmo files.

            --cpcmx SOLVENT
                extended conduction-like polarizable continuum solvation model (CPCM-X),
                available solvents are all solvents included in the Minnesota Solvation Database.
            ```

        Args:
            water (str, optional) : water solvation model (choose 'gbsa' or 'alpb')
                alpb: ALPB solvation model (Analytical Linearized Poisson-Boltzmann).
                gbsa: generalized Born (GB) model with Surface Area contributions.

        Returns:
            SimpleNamespace(PE(total energy in kcal/mol), charges, wbo) 
        """

        with tempfile.TemporaryDirectory() as temp_dir: # tmpdir is a string
            workdir = Path(temp_dir)
            
            geometry_input_path = workdir / 'geometry.xyz'
            xtbout_path = workdir / 'xtbout.json'
            wbo_path = workdir / 'wbo'
            geometry_output_path = workdir / 'xtbtopo.mol'
            
            with open(geometry_input_path, 'w') as geometry:
                geometry.write(self.to_xyz())
            
            cmd = ['xtb', geometry_input_path.as_posix()]
            options = ['-c', str(self.charge), '--sp', '--gfn', '2', '--json']
            
            if water is not None and isinstance(water, str):
                if water == 'gbsa':
                    options += ['--gbsa', 'H2O']
                    # it does not provide Gsolv contribution to the total energy
                elif water == 'alpb':
                    options += ['--alpb', 'water']
                    # it does not provide Gsolv contribution to the total energy
                elif water == 'cpcmx' and self.is_cpcmx_ready():
                    options += ['--cpcmx', 'water']

            if verbose:
                logger.info(f"singlepoint() {' '.join(cmd+options)}")

            # 'xtbout.json', 'xtbrestart', 'xtbtopo.mol', 'charges', and 'wbo' files will be 
            # created in the current working directory.
            proc = subprocess.run(cmd + options, cwd=temp_dir, capture_output=True, text=True, encoding='utf-8')
            # if proc.returncode == 0:
            #     print("Standard Output:")
            #     print(proc.stdout)
            # else:
            #     print("Error:")
            #     print(proc.stderr)
            
            if proc.returncode == 0:
                if xtbout_path.is_file():
                    with open(xtbout_path, 'r') as f:
                        datadict = json.load(f) # takes the file object as input

                Gsolv = None
                                 
                if water is not None:
                    #  Free Energy contributions:                       [Eh]        [kcal/mol]
                    # -------------------------------------------------------------------------
                    #  solvation free energy (dG_solv):             -0.92587E-03    -0.58099
                    #  gas phase energy (E)                         -0.52068E+01
                    # -------------------------------------------------------------------------
                    #  total free energy (dG)                       -0.52077E+01
                    for line in proc.stdout.splitlines():
                        if 'solvation free energy' in line:
                            m = re.search(r"solvation free energy \(dG_solv\)\:\s+[-+]?\d*\.?\d+E[-+]?\d*\s+(?P<kcalpermol>[-+]?\d*\.?\d+)", line)
                            Gsolv = float(m.group('kcalpermol'))
                
                Wiberg_bond_orders = self.load_wbo(wbo_path)

                return SimpleNamespace(
                    natoms = self.natoms,
                    charge = self.charge,
                    PE = datadict['total energy'] * hartree2kcalpermol,
                    Gsolv = Gsolv, 
                    charges = datadict['partial charges'], 
                    wbo = Wiberg_bond_orders,
                    ) 
        
        # something went wrong if it reaches here          
        return SimpleNamespace()
                        


    def optimize(self, water: str | None = None, verbose: bool = False) -> SimpleNamespace:
        """Optimize geometry.

        Options:
            ```sh
            -c, --chrg INT
              specify molecular charge as INT, overrides .CHRG file and xcontrol option
            -o, --opt [LEVEL]
              call ancopt(3) to perform a geometry optimization, levels from crude, sloppy,
              loose, normal (default), tight, verytight to extreme can be chosen
            --gfn INT
              specify parametrisation of GFN-xTB (default = 2)
            --json
              write xtbout.json file
            ```

        Notes:
            Conda installed xtb has Fortran runtime error when optimizing geometry.
            ```sh
            Fortran runtime errror:
                At line 852 of file ../src/optimizer.f90 (unit = 6, file = 'stdout')
                Fortran runtime error: Missing comma between descriptors
                (1x,"("f7.2"%)")
                            ^
                Error termination.
            ```

        Args:
            water (str, optional) : water solvation model (choose 'gbsa' or 'alpb')
                alpb: ALPB solvation model (Analytical Linearized Poisson-Boltzmann).
                gbsa: generalized Born (GB) model with Surface Area contributions.

        Returns:
            (total energy in kcal/mol, optimized geometry)
        """
        with tempfile.TemporaryDirectory() as temp_dir: # tmpdir is a string
            workdir = Path(temp_dir)
            
            geometry_input_path = workdir / 'geometry.xyz'
            xtbout_path = workdir / 'xtbout.json'
            geometry_output_path = workdir / 'xtbopt.xyz'
            wbo_path = workdir / 'wbo'
            
            with open(geometry_input_path, 'w') as geometry:
                geometry.write(self.to_xyz())

            cmd = ['xtb', geometry_input_path.as_posix()]
            options = ['-c', str(self.charge), '-o', 'normal', '--gfn', '2', '--json']

            if water is not None and isinstance(water, str):
                if water == 'gbsa':
                    options += ['--gbsa', 'H2O']
                elif water == 'alpb':
                    options += ['--alpb', 'water']
                elif water == 'cpcmx':
                    logger.warning('optimize with --cpcmx option is not implemented in xtb yet')

            if verbose:
                logger.info(f"optimize() {' '.join(cmd+options)}")

            proc = subprocess.run(cmd + options, cwd=temp_dir, capture_output=True, text=True, encoding='utf-8')

            if proc.returncode == 0 and xtbout_path.is_file():
                with open(xtbout_path, 'r') as f:
                    datadict = json.load(f) # takes the file object as input
                
                Wiberg_bond_orders = self.load_wbo(wbo_path)
                rdmol_opt = self.load_xyz(geometry_output_path)
            
                return SimpleNamespace(
                        natoms = self.natoms,
                        charge = self.charge,
                        PE = datadict['total energy'] * hartree2kcalpermol,
                        charges = datadict['partial charges'],
                        wbo = Wiberg_bond_orders,
                        geometry = rdmol_opt,
                )

        # something went wrong if it reaches here
        return SimpleNamespace()
    

    def esp(self, water: str | None = None, verbose: bool = False) -> None:
        """Calculate electrostatic potential
        
        Example:
            v = py3Dmol.view()
            v.addVolumetricData(dt,
                                "cube.gz", {
                                    'isoval': 0.005,
                                    'smoothness': 2,
                                    'opacity':.9,
                                    'voldata': esp,
                                    'volformat': 'cube.gz',
                                    'volscheme': {
                                        'gradient':'rwb',
                                        'min':-.1,
                                        'max':.1,
                                        }
                                    });
            v.addModel(dt,'cube')
            v.setStyle({'stick':{}})
            v.zoomTo()
            v.show()
        """
        with tempfile.TemporaryDirectory() as temp_dir: # tmpdir is a string
            workdir = Path(temp_dir)
            if verbose:
                logger.info(f'xtb.optimize workdir= {temp_dir}')
            
            geometry_input_path = workdir / 'geometry.xyz'
            xtb_esp_dat = workdir / 'xtb_esp_dat'

            with open(geometry_input_path, 'w') as geometry:
                geometry.write(self.to_xyz())

            cmd = ['xtb', geometry_input_path.as_posix()]

            options = ['--esp', '--gfn', '2']

            if water is not None and isinstance(water, str):
                if water == 'gbsa':
                    options += ['--gbsa', 'H2O']
                elif water == 'alpb':
                    options += ['--alpb', 'water']

            proc = subprocess.run(cmd + options, cwd=temp_dir, capture_output=True, text=True, encoding='utf-8')
            # output files: xtb_esp.cosmo, xtb_esp.dat, xtb_esp_profile.dat

            if proc.returncode == 0 and xtb_esp_dat.is_file():
                with open(xtb_esp_dat, 'r') as f:
                    pass
        
        return None



from xml.etree.ElementTree import Element, SubElement, Comment, tostring
from xml.dom import minidom

root= Element('Brenk2008')
root.append(Comment('Table S1. Definition of unwanted groups'))

f=open('Brenk_et_al-2008-ChemMedChem-SI.txt','rt')
count= 0
for line in f :
	c=line.strip().split('\t')
	count += 1
	entry= SubElement (root, 'group')
	entry.set('name', '('+str(count)+') '+c[0].strip())
	node= SubElement (entry, 'SMARTS')
	node.text= c[1].strip()
f.close()


g=open('Brenk2008.xml','wt')
coarse= tostring(root,'utf-8')
g.write( minidom.parseString( coarse ).toprettyxml(indent="  ") )



# In ASE, the default energy unit is eV (electron volt).
# In xtb, the default energy unit it Eh (hartree).

# CODATA 2018 energy conversion factor

hartree2ev = 27.211386245988
hartree2kcalpermol = 627.50947337481
ev2kcalpermol = 23.060547830619026

# In mendeleev, the default length unit is pm.
# picometer (e-12) to angstrom (e-10)
pm2angstrom = 0.01



import rdworks
import tempfile
import copy
import math

from rdkit import Chem
from rdworks import Mol, MolLibr
from pathlib import Path


datadir = Path(__file__).parent.resolve() / "data"

# python >=3.12 raises SyntaxWarning: invalid escape sequence
# To address this warning in general, we can make the string literal a raw string literal r"...". 
# Raw string literals do not process escape sequences. 
# For example, r"\n" is treated simply as the characters \ and n and not as a newline escape sequence.
drug_smiles = [
    "Fc1cc(c(F)cc1F)C[C@@H](N)CC(=O)N3Cc2nnc(n2CC3)C(F)(F)F", # [0]
    r"O=C(O[C@@H]1[C@H]3C(=C/[C@H](C)C1)\C=C/[C@@H]([C@@H]3CC[C@H]2OC(=O)C[C@H](O)C2)C)C(C)(C)CC",
    "C[C@@H](C(OC(C)C)=O)N[P@](OC[C@@H]1[C@H]([C@@](F)([C@@H](O1)N2C=CC(NC2=O)=O)C)O)(OC3=CC=CC=C3)=O",
    "C1CNC[C@H]([C@@H]1C2=CC=C(C=C2)F)COC3=CC4=C(C=C3)OCO4",
    "CC1=C(C=NO1)C(=O)NC2=CC=C(C=C2)C(F)(F)F",
    "CN1[C@@H]2CCC[C@H]1CC(C2)NC(=O)C3=NN(C4=CC=CC=C43)C", # [5] - Granisetron
    "CCCN1C[C@@H](C[C@H]2[C@H]1CC3=CNC4=CC=CC2=C34)CSC",
    "CCC1=C(NC2=C1C(=O)C(CC2)CN3CCOCC3)C", # [7] Molidone
    r"C[C@H]1/C=C/C=C(\C(=O)NC2=C(C(=C3C(=C2O)C(=C(C4=C3C(=O)[C@](O4)(O/C=C/[C@@H]([C@H]([C@H]([C@@H]([C@@H]([C@@H]([C@H]1O)C)O)C)OC(=O)C)C)OC)C)C)O)O)/C=N/N5CCN(CC5)C)/C",
    r"C=CC1=C(N2[C@@H]([C@@H](C2=O)NC(=O)/C(=N\O)/C3=CSC(=N3)N)SC1)C(=O)O",
    "CC1=C(N=CN1)CSCCNC(=NC)NC#N", # [10] - Cimetidine
    """C1=C(N=C(S1)N=C(N)N)CSCC/C(=N/S(=O)(=O)N)/N""",
    "C1CC(CCC1C2=CC=C(C=C2)Cl)C3=C(C4=CC=CC=C4C(=O)C3=O)O",
    "CN(CC/C=C1C2=CC=CC=C2SC3=C/1C=C(Cl)C=C3)C",
    "CN(C)CCCN1C2=CC=CC=C2CCC3=C1C=C(C=C3)Cl",
    "CN1CCCC(C1)CC2C3=CC=CC=C3SC4=CC=CC=C24", # [15] - Methixene
    "CCN(CC)C(C)CN1C2=CC=CC=C2SC3=CC=CC=C31",
    "CC(=O)OC1=CC=CC=C1C(=O)O",
    "C1=CC(=C(C=C1F)F)C(CN2C=NC=N2)(CN3C=NC=N3)O",
    "CC(=O)NC[C@H]1CN(C(=O)O1)C2=CC(=C(C=C2)N3CCOCC3)F", # [19]
    ]

drug_names = [
    "Sitagliptin", "Simvastatin", "Sofosbuvir", "Paroxetine", "Leflunomide",
    "Granisetron", "Pergolide", "Molindone", "Rifampin", "Cefdinir",
    "Cimetidine", "Famotidine", "Atovaquone", "Chlorprothixene", "Clomipramine",
    "Methixene",  "Ethopropazine", "Aspirin", "Fluconazole", "Linezolid",
    ]



def test_init_mollibr():
    libr = MolLibr(drug_smiles[:5], drug_names[:5])
    assert libr.count() == 5
    libr = MolLibr([Chem.MolFromSmiles(_) for _ in drug_smiles[5:10]], drug_names[5:10])
    assert libr.count() == 5
    libr = MolLibr([Mol(smi,name) for smi,name in zip(drug_smiles[10:15], drug_names[10:15])])
    assert libr.count() == 5


def test_operators():
    libr = MolLibr(drug_smiles, drug_names)
    assert libr.count() == 20

    # other library has 5 overlapping molecules
    other = MolLibr(drug_smiles[5:10], drug_names[5:10])
    assert other.count() == 5

    # index or slice and equality
    assert libr[10] == Mol("CC1=C(N=CN1)CSCCNC(=NC)NC#N")
    assert libr[5:10] == other

    assert (libr + other).count() == 25 # appended, 5 duplicates
    assert (libr - other).count() == 15
    assert (libr & other).count() == 5
    assert libr.count() == 20 # libr object is unchanged.

    # libr will be changed by +=, -=, &= operators
    libr += other
    assert libr.count() == 25 # appended, 5 duplicates

    libr -= other
    assert libr.count() == 15 # now libr has no overlap with other

    libr &= other 
    assert libr.count() == 0 # previous operator removed common molecules


def test_copy():
    libr1 = MolLibr(drug_smiles[:5], drug_names[:5])
    for i in range(5):
        libr1.libr[i].rdmol.SetProp("_Name", f"_Name_{i}")
        libr1.libr[i].rdmol.SetProp("Property", f"Property_{i}")
    libr2 = copy.deepcopy(libr1) # copied
    for i in range(5):
        assert libr2.libr[i].rdmol.GetProp("_Name") == f"_Name_{i}"
        assert libr2.libr[i].rdmol.GetProp("Property") == f"Property_{i}"
        assert libr1.libr[i].smiles == libr2.libr[i].smiles


def test_unique():
    libr = MolLibr(
        drug_smiles[:3] + ["N[C@@H](CC(=O)N1CCN2C(C1)=NN=C2C(F)(F)F)CC1=CC(F)=C(F)C=C1F"], 
        drug_names[:3] + ["Januvia"])
    libr_unique = libr.unique()
    assert libr_unique.count() == 3
    assert libr_unique[0].props['aka'] == ['Januvia']

 
def test_merge_csv():
    libr = MolLibr(drug_smiles, drug_names)
    libr = rdworks.merge_csv(libr, datadir / "drugs_20.csv", on='name')
    with tempfile.TemporaryDirectory() as temp_dir: # tmpdir is a string
        workdir = Path(temp_dir)
        libr.to_csv(workdir / "test_merge_csv.csv")
        assert libr.count() == 20


def test_read_smi():
    libr1 = rdworks.read_smi(datadir / "cdk2.smi", progress=False)
    assert libr1.count() == 47, "failed to read .smi file"
    libr2 = rdworks.read_smi(datadir / "cdk2.smi.gz", progress=False)
    assert libr2.count() == 47, "failed to read .smi.gz file"
    assert libr1 == libr2


def test_drop():
    libr = MolLibr(drug_smiles, drug_names, progress=False)
    not_druglike_names = ['Sofosbuvir','Rifampin','Cefdinir','Famotidine','Atovaquone','Chlorprothixene','Methixene','Ethopropazine']
    cnsmpo_compliant_names = ['Sitagliptin','Simvastatin','Paroxetine','Leflunomide','Granisetron','Molindone','Cimetidine','Fluconazole','Linezolid']

    obj = libr.drop()

    obj = libr.drop('ZINC_druglike')
    assert obj.count() == 8
    assert set([_.name for _ in obj]) == set(not_druglike_names)
    
    obj = libr.drop('~ZINC_druglike')       
    assert obj.count() == 12
    assert set([_.name for _ in obj]) == set(drug_names)-set(not_druglike_names)
    
    # Keep CNS compliant compounds
    # Below three drop() functions have the same effect
    # and obj1, obj2, and obj3 should be identical
    obj1 = libr.drop('CNS', invert=True)
    assert obj1.count() == 9
    assert set([_.name for _ in obj1]) == set(cnsmpo_compliant_names)

    obj2 = libr.drop('~CNS')
    assert obj2.count() == 9
    assert set([_.name for _ in obj2]) == set(cnsmpo_compliant_names)
    
    obj3 = libr.drop(datadir / 'cns.xml', invert=True)
    assert obj3.count() == 9
    assert set([_.name for _ in obj3]) == set(cnsmpo_compliant_names)
        

def test_similar():
    libr = MolLibr(drug_smiles, drug_names, progress=False)
    query = Mol('[H][C@@]1(CC[C@]([H])(C2=C(F)C=C(F)C(F)=C2)[C@@]([H])(N)C1)N1CCN2C(C1)=NN=C2C(F)(F)F', 'DB07072')
    assert libr.similar(query, threshold=0.2).count() == 1
    query = libr[15] # Methixene
    sim = libr.similar(query, threshold=0.2)
    sim_expected = ['Pergolide', 'Methixene', 'Ethopropazine']
    sim_names = [_.name for _ in sim]
    assert set(sim_names) == set(sim_expected)

def test_cluster():
    libr = rdworks.read_smi(datadir / "cdk2.smi.gz", progress=False)
    assert libr.count() == 47
    clusters = libr.cluster(threshold=0.3)
    assert isinstance(clusters, list)
    assert len(clusters) == 3


def test_nnp_ready():
    libr = MolLibr(drug_smiles, drug_names)
    libr_subset = libr.nnp_ready('ANI-2x', progress=False)
    assert libr_subset.count() == 19


def test_qed():
    libr = MolLibr(drug_smiles[:3], drug_names[:3]).qed(progress=False)
    assert math.isclose(libr[0].props['MolWt'], 407.318, rel_tol=1.0e-6, abs_tol=1.0e-3)
    assert math.isclose(libr[0].props['QED'], 0.62163, rel_tol=1.0e-6, abs_tol=1.0e-3)
    # calculate all available descriptors:
    # "MolWt", "TPSA", "LogP", "HBA", "HBD", "QED", "LipinskiHBA", "LipinskiHBD",
    # "HAC", "RotBonds", "RingCount", "Hetero", "FCsp3"
    libr = MolLibr(drug_smiles, drug_names).qed(
        properties=[k for k in rdworks.rd_descriptor_f], 
        progress=False)
    

def test_read_sdf():
    libr1 = rdworks.read_sdf(datadir / "cdk2.sdf", progress=False)
    assert libr1.count() == 47, "failed to read .sdf file"
    libr2 = rdworks.read_sdf(datadir / "cdk2.sdf.gz", progress=False)
    assert libr2.count() == 47, "failed to read .sdf.gz file"
    assert libr1 == libr2


def test_read_mae():
    libr = rdworks.read_mae(datadir / "ligprep-SJ506rev-out.mae")
    print(libr.count())


def test_to_csv():
    libr1 = MolLibr(drug_smiles, drug_names, progress=False)
    with tempfile.TemporaryDirectory() as temp_dir: # tmpdir is a string
        workdir = Path(temp_dir)
        libr1.qed(progress=False).to_csv(workdir / "test_to_csv.csv")
        libr2 = rdworks.read_csv(workdir / "test_to_csv.csv", smiles='smiles', name='name', progress=False)
        assert libr1 == libr2


def test_to_smi():
    libr = MolLibr(drug_smiles, drug_names, progress=False)
    with tempfile.TemporaryDirectory() as temp_dir: # tmpdir is a string
        workdir = Path(temp_dir)
        libr.to_smi(workdir / "test_to_smi.smi.gz")
        libr.to_smi(workdir / "test_to_smi.smi")

   
def test_to_sdf():
    libr = MolLibr(drug_smiles, drug_names, progress=False)
    with tempfile.TemporaryDirectory() as temp_dir: # tmpdir is a string
        workdir = Path(temp_dir)
        libr.to_sdf(workdir / "test_to_sdf.sdf.gz")
        libr.to_sdf(workdir / "test_to_sdf.sdf")
        libr.qed().to_sdf(workdir / "test_to_sdf_with_qed.sdf") # QED and other properties should be here
        supp = Chem.SDMolSupplier(workdir / "test_to_sdf_with_qed.sdf")
        for m, mol in zip(supp, libr):
            assert math.isclose(float(m.GetProp('MolWt')), 
                                mol.props['MolWt'], rel_tol=1.0e-6, abs_tol=1.0e-3)
            assert math.isclose(float(m.GetProp('QED')), 
                                mol.props['QED'], rel_tol=1.0e-6, abs_tol=1.0e-3)


def test_to_png():
    libr = MolLibr(drug_smiles, drug_names, progress=False)
    with tempfile.TemporaryDirectory() as temp_dir: # tmpdir is a string
        workdir = Path(temp_dir)
        libr.to_png(workdir / "test_to_png.png")
        libr.to_png(workdir / "test_to_png_with_index.png", atom_index=True)


def test_to_svg():
    libr = MolLibr(drug_smiles, drug_names, progress=False)
    with tempfile.TemporaryDirectory() as temp_dir: # tmpdir is a string
        workdir = Path(temp_dir)
        with open(workdir / "test_to_svg.svg", "w") as svg:
            svg.write(libr.to_svg())
        head = libr.to_svg()[:100]
        assert head.startswith('<?xml') and ("<svg" in head)


def test_expand_rgroup():
    X = ["[*]C#N", "[*]C(O)=O", "[*]CO", "[*]COC", "[*]C(NC)=O", "[*]CNC(C)=O", "[*]CC=C", "[*][H]" ] # (8)
    Y = ["[*][H]", "[*]O", "[*]OC", "[*]CC(F)(F)F", "[*]OCCOC"] # (5)
    core = "[*:1]-c1ccc2ccn(-[*:2])c2c1"
    libr = rdworks.expand_rgroup(core=core, r={1:X, 2:Y}, prefix='RGX', progress=False)
    assert libr.count() == 40 # 8x5


def test_scaffold_tree():
    libr = MolLibr(drug_smiles[:4], drug_names[:4])
    with tempfile.TemporaryDirectory() as temp_dir: # tmpdir is a string
        workdir = Path(temp_dir)
        for mol in libr:
            adhoc_libr = MolLibr(rdworks.scaffold_tree(mol.rdmol)).rename(prefix=mol.name)
            adhoc_libr.to_png(workdir / f'unittest_84_{mol.name}.png')


def test_MatchedSeries():
    # https://greglandrum.github.io/rdkit-blog/posts/2023-01-09-rgd-tutorial.html
    X = ["[*]C#N", "[*]C(O)=O", "[*]CO", "[*]COC", "[*]C(NC)=O", "[*]CNC(C)=O", "[*]CC=C", "[*][H]" ] # (8)
    Y = ["[*][H]", "[*]O", "[*]OC", "[*]CC(F)(F)F", "[*]OCCOC"] # (5)
    core = "[*:1]-c1ccc2ccn(-[*:2])c2c1"
    libr = rdworks.expand_rgroup(core=core, r={1:X, 2:Y}, prefix='RGX', progress=False)
    series = rdworks.MatchedSeries(libr, sort_props=['QED','HAC'])
    assert series.count() == 10


import numpy as np
import math
import pytest
import pathlib
import tempfile

from rdworks import State, StateEnsemble, StateNetwork


@pytest.fixture(scope='module') # Runs once for every test module (file).
def prepared_state():
    smiles = 'c1ccc(CNc2ncnc3ccccc23)cc1'
    st = State(smiles=smiles)
    return st


def test_site(prepared_state):
    """Ionizable site"""
    st = prepared_state
    assert st.site_info() == [
        ('N', 5, 0, True, True),
        ('N', 7, 0, True, False), 
        ('N', 9, 0, True, False), 
        ]
    # SMILES: c1ccc(CNc2ncnc3ccccc23)cc1
    # Formal charge: 0
    # Origin: None
    # Transformation: None
    # Ionizable sites:
    # - atom_idx=  5, atom=  N, q= +0, hs= 1, pr= 1, de= 1, acid_base= B:A:A, name= Amine:Amide:Amide vinylogue
    # - atom_idx=  7, atom=  N, q= +0, hs= 0, pr= 1, de= 0, acid_base= B, name= Aza-aromatics
    # - atom_idx=  9, atom=  N, q= +0, hs= 0, pr= 1, de= 0, acid_base= B, name= Aza-aromatics
    
    smiles = 'C1=Nc2ccccc2C(N=Cc2ccccc2)N1'
    st = State(smiles=smiles)
    assert st.site_info() == [
        ('N', 1, 0, True, False), 
        ('N', 9, 0, True, False), 
        ('N', 17, 0, True, True)]
    # SMILES: C1=Nc2ccccc2C(N=Cc2ccccc2)N1
    # Formal charge: 0
    # Origin: None
    # Transformation: None
    # Ionizable sites:
    # - atom_idx=  1, atom=  N, q= +0, hs= 0, pr= 1, de= 0, acid_base= B, name= Imine
    # - atom_idx=  9, atom=  N, q= +0, hs= 0, pr= 1, de= 0, acid_base= B, name= Imine
    # - atom_idx= 17, atom=  N, q= +0, hs= 1, pr= 1, de= 1, acid_base= A, name= Amide


def test_tautomers():
    smiles = 'c1ccc(CNc2ncnc3ccccc23)cc1'
    st1 = State(smiles=smiles, tautomer_rule='rdkit')
    se1 = StateEnsemble(st1.get_tautomers())
    assert se1.size() == 2
    st2 = State(smiles=smiles, tautomer_rule='comprehensive')
    se2 = StateEnsemble(st2.get_tautomers()) 
    assert se2.size() == 20


def test_protonate(prepared_state):
    st = prepared_state
    ps = st.get_protonated(atom_idx=9)
    assert len(ps) == 1
    assert ps[0].site_info() == [
        ('N', 5, 0, True, True),
        ('N', 7, 0, True, False), 
        ('N', 9, 1, False, True),
        ]
    # SMILES: c1ccc(CNc2nc[nH+]c3ccccc23)cc1
    # Formal charge: 1
    # Origin: c1ccc(CNc2ncnc3ccccc23)cc1
    # Transformation: +H
    # Ionizable sites:
    # - atom_idx=  5, atom=  N, q= +0, hs= 1, pr= 1, de= 1, acid_base= B:A:A, name= Amine:Amide:Amide vinylogue
    # - atom_idx=  7, atom=  N, q= +0, hs= 0, pr= 1, de= 0, acid_base= B, name= Aza-aromatics
    # - atom_idx=  9, atom=  N, q= +1, hs= 1, pr= 0, de= 1, acid_base= A, name= Aza-aromatics

    ps = st.get_protonated(site_idx=2)
    assert len(ps) == 1
    assert ps[0].site_info() == [
        ('N', 5, 0, True, True),
        ('N', 7, 0, True, False), 
        ('N', 9, 1, False, True),
        ]

    se = StateEnsemble(st.get_protonated())
    assert se.size() == 3
    results = [_.smiles for _ in se]
    expected = ['c1ccc(C[NH2+]c2ncnc3ccccc23)cc1',
                'c1ccc(CNc2[nH+]cnc3ccccc23)cc1',
                'c1ccc(CNc2nc[nH+]c3ccccc23)cc1'
                ]
    assert set(results) == set(expected)


def test_deprotonate(prepared_state):
    st = prepared_state
    des = st.get_deprotonated(atom_idx=5)
    assert len(des) == 1
    assert des[0].site_info() == [
        ('N', 5, -1, True, False),
        ('N', 7, 0, True, False), 
        ('N', 9, 0, True, False),
    ]

    des = st.get_deprotonated(site_idx=0)
    assert len(des) == 1
    assert des[0].site_info() == [
        ('N', 5, -1, True, False),
        ('N', 7, 0, True, False), 
        ('N', 9, 0, True, False),
    ]

    des = st.get_deprotonated(atom_idx=7)
    assert len(des) == 0

    se = StateEnsemble(st.get_deprotonated())
    assert se.size() == 1
    results = [_.smiles for _ in se]
    expected = ['c1ccc(C[N-]c2ncnc3ccccc23)cc1']
    assert set(results) == set(expected)


# def test_sn():
#     smiles = 'c1ccc(CNc2ncnc3ccccc23)cc1'
#     sn1 = StateNetwork()
#     sn1.build(smiles=smiles, tautomer_rule=None)
#     assert len(sn1.visited_states) == 11
#     assert len(sn1.graph.nodes()) == 11
#     sn2 = StateNetwork()
#     sn2.build(smiles=smiles, tautomer_rule='rdkit')
#     assert len(sn2.visited_states) == 33
#     assert len(sn2.graph.nodes()) == 33
#     sn3 = StateNetwork()
#     sn3.build(smiles=smiles, tautomer_rule='comprehensive')
#     assert len(sn3.visited_states) == 183
#     assert len(sn3.graph.nodes()) == 183
    

def test_unipka_workflow():
    smiles = 'c1ccc(CNc2ncnc3ccccc23)cc1'
    sn = StateNetwork()
    sn.build(smiles=smiles, max_formal_charge=3, tautomer_rule=None)
    assert len(sn.visited_states) == 12
    assert len(sn.graph.nodes()) == 12

    sn.info()
    sn.get_initial_state().info()
    sn.get_states().info()

    assert sn.get_num_nodes() == sn.get_states().size()

    # calculated from Uni-pKa
    LN10 = math.log(10)
    TRANSLATE_PH = 6.504894871171601
    # Uni-pka model specific variable for pH dependent deltaG
    # Training might be conducted with a dataset in which raw pKa values
    # were subtracted by the mean value (TRANSLATE_PH).
    FE = np.array([-6.025253772735596, -2.9201512336730957, -2.7405877113342285, 
          -2.9639060497283936, 7.656927108764648, 19.67357063293457, 
          21.269811630249023, 11.911577224731445, 7.5623698234558105, 
          10.144123077392578, 21.36874008178711, 12.132856369018555])
    
    # micro-pKa
    micro = sn.micro_pKa(FE, beta=1.0)
    micro_pKa = {}
    for k, v in micro.items():
        # translation is necessary!!
        micro_pKa[k] = (np.array(v)/LN10 + TRANSLATE_PH).tolist()
    expected_micro_pKa = {
        5: [5.263056256779453, 12.37074708270436], 
        7: [5.190629233663213], 
        9: [5.280577863433882],
        }
    assert micro_pKa.keys() == expected_micro_pKa.keys()
    for k,v in micro_pKa.items():
        for x, y in zip(v, expected_micro_pKa[k]):
            assert math.isclose(x, y)

    # macro-pKa
    macro = sn.macro_pKa(FE, beta=1.0)
    expected_macro_pKa = [5.248700101444452, 12.3693201962264]
    macro_pKa = (np.array(macro)/LN10 + TRANSLATE_PH).tolist()
    for x, y in zip(macro_pKa, expected_macro_pKa):
        assert math.isclose(x, y)

    # population
    pH = np.array([1.2, 7.4, 14.0]) - TRANSLATE_PH
    p = sn.population(FE, pH, C=LN10, beta=1.0)
    expected_p = [
        [2.6935020358203173e-05, 0.983819163640424, 0.027227135908720494], 
        [0.24360336373466238, 0.005614114302596941, 3.9027248338257366e-11], 
        [0.20356346464419176, 0.0046913496629296765, 3.2612529504798934e-11], 
        [0.25449881002178604, 0.005865212152380097, 4.077278781474373e-11], 
        [1.5251999407276225e-16, 8.829269017020774e-06, 0.9727728271965177], 
        [7.57267676610138e-06, 1.1011522978789083e-13, 1.9227998109993388e-28], 
        [1.534655014453089e-06, 2.2315608440876436e-14, 3.89668602382293e-29], 
        [0.01779263728460303, 2.5872494015551734e-10, 4.517778939299372e-25], 
        [3.382885964170599e-11, 1.2356211340110039e-06, 3.419574022416676e-08], 
        [2.5588578837852876e-12, 9.346395100583787e-08, 2.5866091967405752e-09], 
        [0.2805056112545945, 2.5735909639257094e-15, 1.1288239220348465e-36], 
        [7.067163565804634e-08, 1.6287075615586162e-09, 1.1322173195857867e-17]
        ]
   
    for x, y in zip(p, expected_p):
        for xz, yz in zip(x, y):
            assert math.isclose(xz, yz, abs_tol=1e-6)

    print("\n")
    for k, st in enumerate(sn.visited_states):
        print(f"{k:2} {st.smiles:50} {FE[k]:8.3f} {p[k]}")

    # population chart
    pH = np.linspace(0, 14, 60)
    p = sn.population(FE, pH=pH-TRANSLATE_PH, C=LN10, beta=1.0)

    chart, se = sn.population_chart(FE,
                               pH=pH-TRANSLATE_PH, 
                               ignore_below=0.05,
                               pH_label=pH, 
                               C=LN10, 
                               beta=1.0,
                               width=600,
                               height=400)
    se.info()
    
    with tempfile.TemporaryDirectory() as temp_dir: # tmpdir is a string
        workdir = pathlib.Path(__file__).parent.resolve() / "outfiles"
        # workdir = pathlib.Path(temp_dir)
        with open(workdir / 'microstate_population_plot.html', 'w') as f:
            chart_json = chart.to_json()
            html = f"""<!DOCTYPE html>
    <html>
        <head>
            <title>Altair Chart</title>
            <script src="https://cdn.jsdelivr.net/npm/vega@5"></script>
            <script src="https://cdn.jsdelivr.net/npm/vega-lite@5"></script>
            <script src="https://cdn.jsdelivr.net/npm/vega-embed@6"></script>
        </head>
        <body>
            <div id="chart-container"></div>
            <script type="text/javascript">
                var spec = JSON.parse(`{chart_json}`);
                vegaEmbed('#chart-container', spec).then(
                    result => console.log(result)).catch(console.warn);
            </script>
        </body>
    </html>"""
            f.write(html)



import math
import itertools
import logging
import importlib.resources
import itertools
import numpy as np
import pandas as pd
import networkx as nx
import copy
import altair as alt

from collections import deque, defaultdict
from typing import Self, Iterator
from dataclasses import dataclass
from types import SimpleNamespace
from pathlib import Path

from rdkit import Chem

from rdworks import Conf, Mol
from rdworks.xtb.wrapper import GFN2xTB
from rdworks.tautomerism import ComprehensiveTautomers, RdkTautomers


logger = logging.getLogger(__name__)


# adapted from https://github.com/dptech-corp/Uni-pKa/enumerator
smarts_path = importlib.resources.files('rdworks.predefined.ionized')
AcidBasePatterns = pd.read_csv(smarts_path / 'smarts_pattern.csv')
AcidBasePatternsSimple = pd.read_csv(smarts_path / 'simple_smarts_pattern.csv')
UnreasonablePatterns = list(map(Chem.MolFromSmarts, [
    "[#6X5]",
    "[#7X5]",
    "[#8X4]",
    "[*r]=[*r]=[*r]",
    "[#1]-[*+1]~[*-1]",
    "[#1]-[*+1]=,:[*]-,:[*-1]",
    "[#1]-[*+1]-,:[*]=,:[*-1]",
    "[*+2]",
    "[*-2]",
    "[#1]-[#8+1].[#8-1,#7-1,#6-1]",
    "[#1]-[#7+1,#8+1].[#7-1,#6-1]",
    "[#1]-[#8+1].[#8-1,#6-1]",
    "[#1]-[#7+1].[#8-1]-[C](-[C,#1])(-[C,#1])",
    # "[#6;!$([#6]-,:[*]=,:[*]);!$([#6]-,:[#7,#8,#16])]=[C](-[O,N,S]-[#1])",
    # "[#6]-,=[C](-[O,N,S])(-[O,N,S]-[#1])",
    "[OX1]=[C]-[OH2+1]",
    "[NX1,NX2H1,NX3H2]=[C]-[O]-[H]",
    "[#6-1]=[*]-[*]",
    "[cX2-1]",
    "[N+1](=O)-[O]-[H]"
]))


def Boltzmann_weighted_average(energies: list[float], beta: float=1.0) -> float:
    energies = np.array(energies)
    relative_energies = energies - np.min(energies)
    boltzmann_factors = np.exp(-relative_energies * beta)
    Z = np.sum(boltzmann_factors)
    weights = boltzmann_factors / Z 
    return float(np.dot(weights, energies))


@dataclass
class IonizableSite:
    """(de)protonation site information"""
    atom_idx: int
    atom: str
    hs: int # number of H attached to the atom
    q: int # formal charge of the atom
    pr: bool # can be protonated?
    de: bool # can be deprotonated?
    name: str
    acid_base: str
    

class State:
    def __init__(self, 
                 smiles: str,
                 origin: str | None = None, 
                 transformation: str | None = None,
                 min_formal_charge: int = -2,
                 max_formal_charge: int = +2,
                 min_atomic_charge: int = -1,
                 max_atomic_charge: int = +1,
                 protomer_rule: str = 'default',
                 tautomer_rule: str | None = None) -> None:
        """Molecular state.

        Args:
            smiles (str): SMILES
            origin (str | None, optional): original SMILES before tautomerization or ionization. Defaults to None.
            transformation (str | None, optional): Tautomer, +H, -H, or None. Defaults to None.
            min_formal_charge (int, optional): min formal charge. Defaults to -2.
            max_formal_charge (int, optional): max formal charge. Defaults to +2.
            min_atomic_charge (int, optional): min atomic charge. Defaults to -1.
            max_atomic_charge (int, optional): max atomic charge. Defaults to +1.
            protomer_rule (str, optional): 
                Ioniziation patterns ('default' or 'simple').
                Defaults to 'default'.
            tautomer_rule (str, optional): 
                Tautomerization patterns ('rdkit' or 'comprehensive'). 
                Defaults to None.
        """
        self.smiles = smiles
        self.origin = origin # parent or origin
        self.transformation = transformation  # how this state is generated from origin
        self.min_formal_charge = min_formal_charge
        self.max_formal_charge = max_formal_charge
        self.min_atomic_charge = min_atomic_charge
        self.max_atomic_charge = max_atomic_charge
        self.protomer_rule = protomer_rule
        self.tautomer_rule = tautomer_rule

        self.rdmol = None
        self.rdmolH = None
        self.sites = []
        self.charge = None
        self.energy = None
        self.update()


    def __str__(self) -> str:
        """String representation.

        Returns:
            str: short description of the state.
        """
        return f"State(smiles={self.smiles}, sites={self.sites}, transformation={self.transformation}, origin={self.origin})"


    def __eq__(self, other: Self) -> bool:
        """Operator `==`."""
        if isinstance(other, State):
            return self.smiles == other.smiles
        return False
    
    
    def copy(self) -> Self:
        return copy.deepcopy(self)


    def update(self) -> None:
        self.rdmol = Chem.MolFromSmiles(self.smiles)
        self.rdmolH = Chem.AddHs(self.rdmol)
        self.find_ionizable_sites()
        self.charge = Chem.GetFormalCharge(self.rdmol)


    def info(self, index: int | None = None) -> None:
        if isinstance(index, int):
            serial = f'[{index:2}] '
        else:
            serial = ''
        print(f"{serial}SMILES: {self.smiles}")
        print(f"{serial}Origin: {self.origin}")
        print(f"{serial}Charge: {self.charge}")
        print(f"{serial}Energy: {self.energy}")
        print(f"{serial}Transformation: {self.transformation}")
        print(f"{serial}Ionizable sites:")
        
        for site in self.sites:
            print(f"{serial}    atom_idx= {site.atom_idx:2},", end=" ")
            print(f"atom= {site.atom:>2},", end=" ")
            print(f"q= {site.q:+2}, hs= {site.hs:1},", end=" ")
            print(f"pr= {site.pr:1}, de= {site.de:1},", end=" ")
            print(f"acid_base= {site.acid_base}, name= {site.name}")
        print()


    def hydrogen_count(self, idx: int) -> int:
        atom = self.rdmolH.GetAtomWithIdx(idx)
        hydrogen_count = 0
        if atom.GetAtomicNum() == 1:
            for bond in atom.GetNeighbors()[0].GetBonds():
                neighbor = bond.GetOtherAtom(atom)
                if neighbor.GetAtomicNum() == 1:
                    hydrogen_count += 1
        else:
            for bond in atom.GetBonds():
                neighbor = bond.GetOtherAtom(atom)
                if neighbor.GetAtomicNum() == 1:
                    hydrogen_count += 1
        return hydrogen_count
    

    def site_info(self) -> list[tuple]:
        return [(site.atom, site.atom_idx, site.q, site.pr, site.de) for site in self.sites]       


    def can_be_protonated_at(self, atom_idx:int) -> bool:
        """Check if an atom can potentially be protonated"""
        atom = self.rdmol.GetAtomWithIdx(atom_idx)
        # Check formal charge (negative charge can be protonated)
        if atom.GetFormalCharge() < 0:
            return True
        
        # Check for atoms with lone pairs (N, O, S, P, etc.)
        # that aren't already fully protonated
        atomic_num = atom.GetAtomicNum()
        total_valence = atom.GetTotalValence()
        
        # Common protonatable atoms
        if atomic_num == 7:  # N, O, S
            if total_valence < 4:  # Can form NH4+
                return True
        elif atomic_num in [8, 16]:  # O, S
            if total_valence < 3:  # Can form OH3+ or SH3+
                return True
        
        return False


    def can_be_deprotonated_at(self, atom_idx:int) -> bool:
        """Check if an atom can potentially be deprotonated"""
        atom = self.rdmol.GetAtomWithIdx(atom_idx)
        # Check if atom has a positive formal charge (can lose H+)
        if atom.GetFormalCharge() > 0:
            return True
        
        # Check if atom has hydrogens that can be removed
        if atom.GetTotalNumHs() == 0:
            return False
        
        # Common deprotonatable atoms with acidic hydrogens
        if atom.GetAtomicNum() in [7, 8, 15, 16]:  # N, O, P, S
            return True
        
        return False
    

    def find_ionizable_sites(self) -> None:
        if self.protomer_rule == 'simple':
            template = AcidBasePatternsSimple
        elif self.protomer_rule == 'default':
            template = AcidBasePatterns
        else:
            template = AcidBasePatterns
        for idx, name, smarts, index, acid_base in template.itertuples():
            pattern = Chem.MolFromSmarts(smarts)
            match = self.rdmolH.GetSubstructMatches(pattern)
            if len(match) == 0:
                continue
            else:
                index = int(index)
                for m in match:
                    atom_idx = m[index]
                    at = self.rdmol.GetAtomWithIdx(atom_idx)
                    atom = at.GetSymbol()
                    hs = self.hydrogen_count(atom_idx)
                    q = at.GetFormalCharge()
                    pr = self.can_be_protonated_at(atom_idx)
                    de = self.can_be_deprotonated_at(atom_idx)
                    site = IonizableSite(atom_idx=atom_idx, 
                                        atom=atom,
                                        hs=hs,
                                        q=q,
                                        name=name,
                                        acid_base=acid_base,
                                        pr=pr,
                                        de=de)
                    exist = False
                    for _ in self.sites:
                        if _.atom_idx == site.atom_idx:
                            exist = True
                            _.acid_base += f':{site.acid_base}'
                            _.name += f':{site.name}'
                    if not exist:
                        self.sites.append(site)
        self.sites = sorted(self.sites, key=lambda x: x.atom_idx)


    def ionize(self, idx: int, mode: str) -> None:
        rwmol = Chem.RWMol(self.rdmol)
        atom = rwmol.GetAtomWithIdx(idx)
        if mode == "a2b":
            if atom.GetAtomicNum() == 1:
                atom_X = atom.GetNeighbors()[0] # only one
                charge = atom_X.GetFormalCharge() -1
                atom_X.SetFormalCharge(charge) # <-- change formal charge
                rwmol.RemoveAtom(idx) # remove the H atom
                rwmol.RemoveBond(idx, atom_X.GetIdx()) # remove the bond    
                ionized = rwmol.GetMol()         
            else:
                charge = atom.GetFormalCharge() -1
                numH = atom.GetTotalNumHs() -1
                atom.SetFormalCharge(charge) # <-- change formal charge
                atom.SetNumExplicitHs(numH) # <-- remove one H
                atom.UpdatePropertyCache() # <-- update the property cache
                ionized = Chem.AddHs(rwmol)
        
        elif mode == "b2a":
            charge = atom.GetFormalCharge() + 1
            atom.SetFormalCharge(charge) # <-- change formal charge
            numH = atom.GetNumExplicitHs() + 1
            atom.SetNumExplicitHs(numH) # <-- add one H
            ionized = Chem.AddHs(rwmol)
            # Add hydrogens, specifying onlyOnAtoms to target the desired atom
            # explicitOnly=True ensures only explicit Hs are added, not implicit ones
            # ionized = Chem.AddHs(mw, explicitOnly=True, onlyOnAtoms=[idx])

        Chem.SanitizeMol(ionized)

        rdmol = Chem.MolFromSmiles(Chem.MolToSmiles(ionized, canonical=False))
        rdmolH = Chem.AddHs(rdmol)
        smiles = Chem.CanonSmiles(Chem.MolToSmiles(Chem.RemoveHs(rdmolH)))

        self.smiles = smiles
        self.sites = []
        self.update()


    def get_protonated(self, 
                        atom_idx: int | None = None, 
                        site_idx: int | None = None) -> list[Self]:
        """Make protonated state(s) from the current state.

        All ionizable sites are considered for protonation unless `atom_idx` or `site_idx` is given.

        Args:
            atom_idx (int | None, optional): atom index. Defaults to None.
            site_idx (int | None, optional): site index. Defaults to None.

        Returns:
            list[Self]: list of protonated States.
        """
        states = []

        if self.charge == self.max_formal_charge:
            return states
        
        if isinstance(atom_idx, int):
            for site in self.sites:
                if site.pr and (site.atom_idx == atom_idx):
                    new_state = self.copy()
                    new_state.ionize(site.atom_idx, "b2a")
                    new_state.transformation = '+H'
                    new_state.origin = self.smiles
                    states.append(new_state)        
        elif isinstance(site_idx, int):
            site = self.sites[site_idx]
            if not site.pr:
                return states
            new_state = self.copy()
            new_state.ionize(site.atom_idx, "b2a")
            new_state.transformation = '+H'
            new_state.origin = self.smiles
            states.append(new_state)
        else:
            for site in self.sites:
                if not site.pr:
                    continue
                new_state = self.copy()
                new_state.ionize(site.atom_idx, "b2a")
                new_state.transformation = '+H'
                new_state.origin = self.smiles
                states.append(new_state)

        return states
    

    def get_deprotonated(self, 
                          atom_idx: int | None = None, 
                          site_idx: int | None = None) -> list[Self]:
        """Make deprotonated state(s) from the current state.

        Args:
            atom_idx (int | None, optional): atom index. Defaults to None.
            site_idx (int | None, optional): site index. Defaults to None.

        Returns:
            list[Self]: list of deprotonated States.
        """
        states = []

        if self.charge == self.min_formal_charge:
            return states
        
        if isinstance(atom_idx, int):
            for site in self.sites:
                if site.de and (site.atom_idx == atom_idx):
                    new_state = self.copy()
                    new_state.ionize(atom_idx, "a2b")
                    new_state.transformation = '-H'
                    new_state.origin = self.smiles
                    states.append(new_state)
        elif isinstance(site_idx, int):
            site = self.sites[site_idx]
            if not site.de:
                return states
            new_state = self.copy()
            new_state.ionize(site.atom_idx, "a2b")
            new_state.transformation = '-H'
            new_state.origin = self.smiles
            states.append(new_state)
        else:
            for site in self.sites:
                if not site.de:
                    continue
                new_state = self.copy()
                new_state.ionize(site.atom_idx, "a2b")
                new_state.transformation = '-H'
                new_state.origin = self.smiles
                states.append(new_state)
        return states


    def get_tautomers(self) -> list[Self]:
        if self.tautomer_rule is None:
            return []
        elif self.tautomer_rule == "rdkit":
            t = RdkTautomers(self.smiles).enumerate()
        elif self.tautomer_rule == "comprehensive":
            t = ComprehensiveTautomers(self.smiles).enumerate()
        else:
            return []
        
        states = []
        for smiles in t.enumerated:
            try:
                assert smiles != self.smiles
                rdmol = Chem.MolFromSmiles(smiles)
                assert rdmol is not None
                charge = Chem.GetFormalCharge(rdmol)
                assert charge == self.charge
                states.append(State(smiles=smiles,
                                    origin=self.smiles,
                                    transformation='Tautomer'))
            except:
                continue
        return states
    


class StateEnsemble:
    def __init__(self, states: list[State] | None = None, transformation: str | None = None) -> None:
        if isinstance(states, list):
            self.states = states
        else:
            self.states = []
        self.maxiter = 10
        if transformation:
            for state in self.states:
                state.transformation = transformation

    def __iter__(self) -> Iterator:
        return iter(self.states)

    def __next__(self) -> State:
        return next(self.states)
    
    def __getitem__(self, index: int | slice) -> State | Self:
        """Operator `[]`"""
        assert self.size() != 0, "StateEnsemble is empty"
        if isinstance(index, slice):
            return StateEnsemble(self.states[index])
        else:
            return self.states[index]

    def __setitem__(self, index: int, state: State) -> Self:
        """Set item."""
        self.states[index] = state
        return self


    def __add__(self, other: State | Self) -> Self:
        """Operator `+`."""
        assert isinstance(other, State | StateEnsemble), "'+' operator expects State or StateEnsemble object"
        new_object = self.copy()
        if isinstance(other, State):
            new_object.states.append(other)
        elif isinstance(other, StateEnsemble):
            new_object.states.extend(other.states)
        return new_object
    

    def __iadd__(self, other: State | Self) -> Self:
        """Operator `+=`."""
        assert isinstance(other, State | StateEnsemble), "'+=' operator expects State or StateEnsemble object"
        if isinstance(other, State):
            self.states.append(other)
        elif isinstance(other, StateEnsemble):
            self.states.extend(other.states)
        return self
    

    def size(self) -> int:
        return len(self.states)
    

    def drop(self) -> Self:
        """Drop duplicate and unreasonable states."""
        U = []
        mask = []
        for state in self.states:
            if state.rdmol is None:
                mask.append(False)
                continue
            if state.smiles in U:
                mask.append(False)
                continue
            reasonable = True
            for pattern in UnreasonablePatterns:
                if len(state.rdmol.GetSubstructMatches(pattern)) > 0:
                    reasonable = False
                    break
            if not reasonable:
                mask.append(False)
                continue
            mask.append(True)
            U.append(state.smiles)  
        self.states = list(itertools.compress(self.states, mask))
        return self
    

    def info(self) -> None:
        for i, state in enumerate(self.states):
            state.info(index=i)
        

class StateNetwork:
    def __init__(self):
        self.graph = nx.DiGraph()
        self.visited_states = []
        self.initial_state = None
    

    def info(self) -> None:
        print(f"StateNetwork - nodes: {self.get_num_nodes()} edeges: {self.get_num_edges()}")


    def size(self) -> int:
        return len(self.visited_states)
    

    def get_num_nodes(self) -> int:
        return len(self.graph.nodes())
    

    def get_num_edges(self) -> int:
        return len(self.graph.edges())
    

    def get_states(self, index: int | slice | None = None) -> StateEnsemble:
        if isinstance(index, slice):
            return StateEnsemble(self.visited_states[index])
        elif isinstance(index, int):
            return StateEnsemble([self.visited_states[index]])
        else: # all
            return StateEnsemble(self.visited_states)


    def get_initial_state(self) -> State:
        return self.initial_state
    

    def generate_neighbors(self, state: State) -> StateEnsemble:
        """Generate all possible neighboring microstates."""
        neighbors = StateEnsemble()
        if state == self.initial_state and isinstance(state.tautomer_rule, str):
            neighbors += StateEnsemble(state.get_tautomers())
        neighbors += StateEnsemble(state.get_protonated())
        neighbors += StateEnsemble(state.get_deprotonated())
        neighbors = neighbors.drop()
        
        return neighbors


    def mcs_index_map(self, other: State) -> dict[int, int]:
        """Mapping atom indices using the maximum common structure (MCS).

        Uses the self.initial_state as reference in mapping `other` State.

        Args:
            other (State): to be mapped state.

        Returns:
            dict: {ref atom index: other atom index, ...}
        """
        mcs = Chem.rdFMCS.FindMCS([self.initial_state.rdmol, other.rdmol], 
                            atomCompare=Chem.rdFMCS.AtomCompare.CompareAny, 
                            bondCompare=Chem.rdFMCS.BondCompare.CompareAny, 
                            completeRingsOnly=True)
        mcs_rdmol = Chem.MolFromSmarts(mcs.smartsString)
        match_1 = self.initial_state.rdmol.GetSubstructMatch(mcs_rdmol)
        match_2 = other.rdmol.GetSubstructMatch(mcs_rdmol)
        return {match_2[i]: match_1[i] for i in range(len(match_1))}


    def build(self, 
              smiles: str,
              origin: str | None = None, 
              transformation: str | None = None,
              min_formal_charge: int = -2,
              max_formal_charge: int = +2,
              min_atomic_charge: int = -1,
              max_atomic_charge: int = +1,
              protomer_rule: str = 'default',
              tautomer_rule: str | None = None,
              verbose: bool = False) -> None:
        """Build the microstate network using BFS from initial state.""" 
        self.initial_state = State(smiles=smiles,
                              protomer_rule=protomer_rule,
                              tautomer_rule=tautomer_rule,
                              min_formal_charge=min_formal_charge,
                              max_formal_charge=max_formal_charge,
                              min_atomic_charge=min_atomic_charge,
                              max_atomic_charge=max_atomic_charge)
        self.initial_state
        # Initialize BFS
        queue = deque([self.initial_state])
        self.visited_states.append(self.initial_state)
        self.graph.add_node(self.initial_state.smiles, 
                            initial=True, 
                            sites=self.initial_state.site_info())
        iter = 0

        while queue:
            iter += 1
            current_state = queue.popleft()
            neighbors = self.generate_neighbors(current_state)
            for neighbor_state in neighbors:
                if neighbor_state.transformation == 'Tautomer' and current_state.charge != neighbor_state.charge:
                    continue
                self.graph.add_edge(current_state.smiles, 
                                    neighbor_state.smiles, 
                                    transformation=neighbor_state.transformation)
                if neighbor_state not in self.visited_states:
                    self.visited_states.append(neighbor_state)
                    imap = self.mcs_index_map(neighbor_state)
                    sites = [(a, imap[i], q, pr, de) for (a, i, q, pr, de) in neighbor_state.site_info()]
                    self.graph.add_node(neighbor_state.smiles, 
                                        initial=False, 
                                        sites=sites)
                    queue.append(neighbor_state)
            if verbose:
                print(f"Iteration {iter:2}: {len(self.visited_states):2} microstates found")
        
        if verbose:
            print(f"\nNetwork construction complete!")
            print(f"Total microstates: {len(self.graph.nodes())}")
            print(f"Total transformations: {len(self.graph.edges())}")


    def micro_pKa(self, FE: np.ndarray, beta: float=1.0) -> dict[int,np.ndarray]:
        """Calculate micro-pKa with provided potential energies.

        Args:
            FE (np.ndarray): free energies of states.
            beta (float, optional): 
                :math:`\\beta = \\frac{1}{kT}` in which :math:`k` is Boltzmann constant 
                and :math:`T` is Temperature in Kelvin unit. 
                Defaults to 1.0.

        Returns:
            dict[int,list[float]]: micro-pKa values for each ionizable site.
        """
        assert FE.shape[0] == len(self.visited_states), "FE should be given for all states"
        for k, st in enumerate(self.visited_states):
            st.energy = FE[k]
        pKa = defaultdict(list)
        for site in self.initial_state.sites:
            group = defaultdict(list)
            for st in self.visited_states:
                for (a, i, q, pr, de) in st.site_info():
                    # ex. [('N', 5, 0, True, True), ...]
                    if i == site.atom_idx:
                        group[q].append(st.energy)
            weighted_mean = {q: Boltzmann_weighted_average(_, beta=beta) for q, _ in sorted(group.items())}
            charges = list(sorted(weighted_mean)) # ex. [-1, 0, +1]
            for (q1, q2) in list(itertools.pairwise(charges)): # ex. [(-1, 0), (0, +1)]
                G_deprotonated = weighted_mean[q1]
                G_protonated = weighted_mean[q2]
                delta_G_deprotonation = G_deprotonated - G_protonated
                pKa[site.atom_idx].append(delta_G_deprotonation)
        
        ordered_pKa = {}
        for k, v in pKa.items():
            ordered_pKa[k] = sorted(v)

        return ordered_pKa
            

    def macro_pKa(self, FE: np.ndarray, beta: float = 1.0) -> np.ndarray:
        """Calculatate macro-pKa with provided potential energies.

        Args:
            FE (np.ndarray): potential energies of states.
            beta (float, optional): 
                :math:`\\beta = \\frac{1}{kT}` in which :math:`k` is Boltzmann constant 
                and :math:`T` is Temperature in Kelvin unit. 
                Defaults to 1.0.

        Returns:
            list[float]: macro-pKa values.
        """
        assert FE.shape[0] == len(self.visited_states), "FE should be given for all states"
        for k, st in enumerate(self.visited_states):
            st.energy = FE[k]
        pKa = []
        for site in self.initial_state.sites:
            group = defaultdict(list)
            for st in self.visited_states:
                for (a, i, q, pr, de) in st.site_info():
                    # ex. [('N', 5, 0, True, True), ...]
                    group[q].append(st.energy)
        weighted_mean = {q: Boltzmann_weighted_average(_, beta=beta) for q, _ in sorted(group.items())}
        charges = list(sorted(weighted_mean)) # ex. [-1, 0, +1]
        for (q1, q2) in list(itertools.pairwise(charges)): # ex. [(-1, 0), (0, +1)]
            G_deprotonated = weighted_mean[q1]
            G_protonated = weighted_mean[q2]
            delta_G_deprotonation = G_deprotonated - G_protonated
            pKa.append(delta_G_deprotonation)
        
        return sorted(pKa)

    
    def population(self, 
                   FE: np.ndarray, 
                   pH: np.ndarray, 
                   C: float, 
                   beta: float = 1.0) -> np.ndarray:
        """Calculate populations with provided potential energies and pH.

        The reference state can be arbitrary and is set to the initial_state here.

        .. math::
            \\Delta G_{i, ref} = PE_{i} - PE_{ref} \\\\
            \\Delta m_{i, ref} = charge_{i} - charge _{ref} \\
            \\Delta G_{i, pH} = \\Delta G_{i, ref} + \\Delta m_{i, ref} C pH \\
            p_{i, pH} = \\frac {exp(-\\beta \\Delta G_{i, pH})}{\\sum_{i} exp(-\\beta \\Delta G_{i, pH})}

        Args:
            FE (np.ndarray): array of potential energies of states.
            pH (np.ndarray): array of pH values.
            C (float): constant for pH-dependent dG calculation.
            beta (float, optional): 
                :math:`\\beta = \\frac{1}{kT}` in which :math:`k` is Boltzmann constant 
                and :math:`T` is Temperature in Kelvin unit. 
                Defaults to 1.0.

        Returns:
            np.ndarray: 
                array of populations with shape of (number of states, number of pH).
        """
        assert FE.shape[0] == len(self.visited_states), "FE should be given for all states"
        ref = None
        for k, st in enumerate(self.visited_states):
            st.energy = FE[k]
            if st == self.initial_state:
                ref = st

        pH = np.array(pH)
        
        dG = []
        for k, st in enumerate(self.visited_states):
            delta_G = FE[k] - ref.energy
            delta_m = st.charge - ref.charge
            dG.append(delta_G + delta_m * C * pH)
        
        dG = np.array(dG)
        Boltzmann_factors = np.exp(-beta * dG)
        Z = np.sum(Boltzmann_factors, axis=0)
        p = Boltzmann_factors/Z
        assert p.shape == (len(self.visited_states), pH.shape[0])
        
        return p


    def population_chart(self, 
                         FE: np.ndarray, 
                         pH: np.ndarray, 
                         C: float, 
                         beta: float = 1.0, 
                         ignore_below: float = 0.01,
                         pH_label: np.ndarray | None = None,
                         width: int = 600,
                         height: int = 400,
                         palette: str = 'tableau10') -> tuple[alt.vegalite.v5.api.LayerChart, StateEnsemble]:
        """Make an Altair plot object for pH-dependent population curve.

        Args:
            FE (np.ndarray): array of potential energies of states.
            pH (np.ndarray): array of pH values.
            C (float): constant for pH-dependent dG calculation.
            beta (float, optional): 
                :math:`\\beta = \\frac{1}{kT}` in which :math:`k` is Boltzmann constant 
                and :math:`T` is Temperature in Kelvin unit. 
                Defaults to 1.0.
            ignore_below (float) : 
                ignore low populated microstates. Defaults to 0.01.
            pH_label (np.ndarray | None, optional): 
                array of pH values for plot. 
                If None, pH is used. Defaults to None.
            width (int, optional): plot width. Defaults to 600.
            height (int, optional): plot height. Defaults to 300.
            palette (str, optional): color palette. Defaults to 'tableau10'.

        Returns:
            alt.vegalite.v5.api.LayerChart: Altair plot object
        """
        
        # calculate populations
        p = self.population(FE, pH, C, beta)
        
        if pH_label is None:
            pH_label = pH
        
        # preparing dataframe and state ensemble
        data = {'Microstate': [], 'pH':[], 'p':[]}
        state_ensemble = StateEnsemble()
        for i, state_populations in enumerate(p):
            if max(state_populations) < ignore_below:
                continue
            state_ensemble += self.get_states(i)
            for j, p in enumerate(state_populations):
                data['Microstate'].append(i)
                data['pH'].append(pH_label[j])
                data['p'].append(p)
        df = pd.DataFrame(data)
        
        # line plot
        lineplot = alt.Chart(df).mark_line().encode(
            x=alt.X('pH:Q', title='pH'),
            y=alt.Y('p:Q', title='Population'),
            color=alt.Color('Microstate:N', scale=alt.Scale(scheme=palette)),
        ).properties(
            width=width,
            height=height)
        
        # data labels
        labels = alt.Chart(df).mark_text(
            align='left',
            dx=5,
            dy=-5
        ).encode(
            x=alt.X('pH', aggregate={'argmax': 'p'}),
            y=alt.Y('p', aggregate={'argmax': 'p'}),
            text='Microstate:N',
            color='Microstate:N'
        )

        chart = (lineplot + labels)
        
        return (chart, state_ensemble)



class QupkakeMicrostates():

    def __init__(self, origin: Mol, calculator: str = 'xTB'):
        self.origin = origin
        self.calculator = calculator
        self.basic_sites = []
        self.acidic_sites = []
        self.states = []
        self.mols = []
        self.reference = None
    

    def enumerate(self) -> None:
        # Qu pKake results must be stored at .confs
        for conf in self.origin:
            pka = conf.props.get('pka', None)
            if pka is None:
                # no protonation/deprotonation sites
                continue
            if isinstance(pka, str) and pka.startswith('tensor'):
                # ex. 'tensor(9.5784)'
                pka = float(pka.replace('tensor(','').replace(')',''))
            if conf.props.get('pka_type') == 'basic':
                self.basic_sites.append(conf.props.get('idx'))
            elif conf.props.get('pka_type') == 'acidic':
                self.acidic_sites.append(conf.props.get('idx'))

        # enumerate protonation/deprotonation sites to generate microstates

        np = len(self.basic_sites)
        nd = len(self.acidic_sites)
        P = [c for n in range(np+1) for c in itertools.combinations(self.basic_sites, n)]
        D = [c for n in range(nd+1) for c in itertools.combinations(self.acidic_sites, n)]
        
        PD = list(itertools.product(P, D))
        
        for (p, d) in PD:
            conf = self.origin.confs[0].copy()
            conf = conf.protonate(p).deprotonate(d).optimize(calculator=self.calculator)
            charge = len(p) - len(d)
            self.states.append(SimpleNamespace(
                charge=charge, 
                protonation_sites=p, 
                deprotonation_sites=d,
                conf=conf,
                smiles=Mol(conf).smiles,
                delta_m=None,
                PE=None))
            
        # sort microstates by ascending charges
        self.states = sorted(self.states, key=lambda x: x.charge)


    @staticmethod
    def Boltzmann_weighted_average(potential_energies: list) -> float:
        """Calculate Boltzmann weighted average potential energy at pH 0.

        Args:
            potential_energies (list): a list of potential energies.

        Returns:
            float: Boltzmann weighted average potential energy.
        """
        kT = 0.001987 * 298.0 # (kcal/mol K), standard condition
        C = math.log(10) * kT
        pe_array = np.array(potential_energies)
        pe = pe_array - min(potential_energies)
        Boltzmann_factors = np.exp(-pe/kT)
        Z = np.sum(Boltzmann_factors)
        p = Boltzmann_factors/Z

        return float(np.dot(p, pe_array))


    def populate(self) -> None:
        for microstate in self.states:
            mol = Mol(microstate.conf).make_confs(n=4).optimize_confs()
            # mol = mol.drop_confs(similar=True, similar_rmsd=0.3, verbose=True)
            # mol = mol.optimize_confs(calculator=calculator)
            # mol = mol.drop_confs(k=10, window=15.0, verbose=True)
            PE = []
            for conf in mol.confs:
                conf = conf.optimize(calculator=self.calculator, verbose=True)
                # GFN2xTB requires 3D coordinates
                # xtb = GFN2xTB(conf.rdmol).singlepoint(water='cpcmx', verbose=True)
                PE.append(conf.potential_energy(calculator=self.calculator))
                # xtb = GFN2xTB(conf.rdmol).singlepoint(verbose=True)
                # SimpleNamespace(
                #             PE = datadict['total energy'] * hartree2kcalpermol,
                #             Gsolv = Gsolv,
                #             charges = datadict['partial charges'],
                #             wbo = Wiberg_bond_orders,
                #             )
            microstate.PE = self.Boltzmann_weighted_average(PE)
            logger.info(f"PE= {PE}")
            logger.info(f"Boltzmann weighted= {microstate.PE}")            
            self.mols.append(mol)

    def get_populations(self, pH: float) -> list[tuple]:
        # set the lowest dG as the reference
        self.reference = self.states[np.argmin([microstate.PE for microstate in self.states])]
        for microstate in self.states:
            microstate.delta_m = microstate.charge - self.reference.charge
        dG = []
        for microstate in self.states:
            dG.append((microstate.PE - self.reference.PE) + microstate.delta_m * C * pH)
        dG = np.array(dG)

        logger.info(f"dG= {dG}")
        kT = 0.001987 * 298.0 # (kcal/mol K), standard condition
        C = math.log(10) * kT
        Boltzmann_factors = np.exp(-dG/kT)
        Z = np.sum(Boltzmann_factors)
        p = Boltzmann_factors/Z
        idx_p = sorted(list(enumerate(p)), key=lambda x: x[1], reverse=True)
        # [(0, p0), (1, p1), ...]

        return idx_p

    def get_ensemble(self) -> list[Mol]:
        return self.mols

    def get_mol(self, idx: int) -> Mol:
        return self.mols[idx]
    
    def count(self) -> int:
        return len(self.states)